// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections.Specialized;
using System.Text;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents the base class for authentication in the EonaCat network library.
    /// </summary>
    internal abstract class AuthenticationBase
    {
        /// <summary>
        /// Gets or sets the collection of authentication parameters.
        /// </summary>
        internal NameValueCollection Parameters;

        /// <summary>
        /// Initializes a new instance of the <see cref="AuthenticationBase"/> class.
        /// </summary>
        /// <param name="scheme">The authentication scheme.</param>
        /// <param name="parameters">The collection of authentication parameters.</param>
        protected AuthenticationBase(AuthenticationSchemes scheme, NameValueCollection parameters)
        {
            Scheme = scheme;
            Parameters = parameters;
        }

        /// <summary>
        /// Gets the algorithm used for authentication.
        /// </summary>
        public string Algorithm => Parameters["algorithm"];

        /// <summary>
        /// Gets the nonce value for authentication.
        /// </summary>
        public string Nonce => Parameters["nonce"];

        /// <summary>
        /// Gets the opaque value for authentication.
        /// </summary>
        public string Opaque => Parameters["opaque"];

        /// <summary>
        /// Gets the quality of protection for authentication.
        /// </summary>
        public string Qop => Parameters["qop"];

        /// <summary>
        /// Gets the realm for authentication.
        /// </summary>
        public string Realm => Parameters["realm"];

        /// <summary>
        /// Gets the authentication scheme.
        /// </summary>
        public AuthenticationSchemes Scheme { get; }

        /// <summary>
        /// Creates a nonce value for authentication.
        /// </summary>
        /// <returns>A string representing the generated nonce value.</returns>
        internal static string CreateNonceValue()
        {
            var buffer = new byte[16];
            var random = new Random();
            random.NextBytes(buffer);

            var result = new StringBuilder(32);
            foreach (var currentByte in buffer)
            {
                result.Append(currentByte.ToString("x2"));
            }

            return result.ToString();
        }

        /// <summary>
        /// Parses the authentication parameters from the specified string value.
        /// </summary>
        /// <param name="value">The string containing authentication parameters.</param>
        /// <returns>A collection of authentication parameters.</returns>
        internal static NameValueCollection ParseParameters(string value)
        {
            var result = new NameValueCollection();
            foreach (var param in value.SplitHeaderValue(','))
            {
                var i = param.IndexOf('=');
                var name = i > 0 ? param.Substring(0, i).Trim() : null;
                var val = i < 0
                          ? param.Trim().Trim('"')
                          : i < param.Length - 1
                            ? param.Substring(i + 1).Trim().Trim('"')
                            : string.Empty;

                result.Add(name, val);
            }

            return result;
        }

        /// <summary>
        /// Gets the authentication string for Basic authentication.
        /// </summary>
        /// <returns>A string representing the Basic authentication.</returns>
        internal abstract string ToBasicString();

        /// <summary>
        /// Gets the authentication string for Digest authentication.
        /// </summary>
        /// <returns>A string representing the Digest authentication.</returns>
        internal abstract string ToDigestString();

        /// <summary>
        /// Returns a string representation of the authentication information.
        /// </summary>
        /// <returns>A string representation of the authentication information.</returns>
        public override string ToString()
        {
            return Scheme == AuthenticationSchemes.Basic
                   ? ToBasicString()
                   : Scheme == AuthenticationSchemes.Digest
                     ? ToDigestString()
                     : string.Empty;
        }
    }
}