// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Runtime.InteropServices;
using System.Runtime.Serialization;
using System.Security.Permissions;
using System.Text;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents a collection of HTTP headers in the EonaCat Network library.
    /// </summary>
    [Serializable]
    [ComVisible(true)]
    public class WebHeaderCollection : NameValueCollection, ISerializable
    {
        private const int BUFFER_SIZE = 65535;
        private static readonly Dictionary<string, HttpHeaderInfo> _headers;
        private readonly bool _internallyUsed;
        private HttpHeaderType _state;

        static WebHeaderCollection()
        {
            _headers = new Dictionary<string, HttpHeaderInfo>(StringComparer.InvariantCultureIgnoreCase)
            {
                { "Accept", new HttpHeaderInfo("Accept", HttpHeaderType.Request | HttpHeaderType.Restricted | HttpHeaderType.MultiValue) },
                { "AcceptCharset", new HttpHeaderInfo("Accept-Charset", HttpHeaderType.Request | HttpHeaderType.MultiValue) },
                { "AcceptEncoding", new HttpHeaderInfo("Accept-Encoding", HttpHeaderType.Request | HttpHeaderType.MultiValue) },
                { "AcceptLanguage", new HttpHeaderInfo("Accept-Language", HttpHeaderType.Request | HttpHeaderType.MultiValue) },
                { "AcceptRanges", new HttpHeaderInfo("Accept-Ranges", HttpHeaderType.Response | HttpHeaderType.MultiValue) },
                { "Age", new HttpHeaderInfo("Age", HttpHeaderType.Response) },
                { "Allow", new HttpHeaderInfo("Allow", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.MultiValue) },
                { "Authorization", new HttpHeaderInfo("Authorization", HttpHeaderType.Request | HttpHeaderType.MultiValue) },
                { "CacheControl", new HttpHeaderInfo("Cache-Control", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.MultiValue) },
                { "Connection", new HttpHeaderInfo("Connection", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.Restricted | HttpHeaderType.MultiValue) },
                { "ContentEncoding", new HttpHeaderInfo("Content-Encoding", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.MultiValue) },
                { "ContentLanguage", new HttpHeaderInfo("Content-Language", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.MultiValue) },
                { "ContentLength", new HttpHeaderInfo("Content-Length", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.Restricted) },
                { "ContentLocation", new HttpHeaderInfo("Content-Location", HttpHeaderType.Request | HttpHeaderType.Response) },
                { "ContentMd5", new HttpHeaderInfo("Content-MD5", HttpHeaderType.Request | HttpHeaderType.Response) },
                { "ContentRange", new HttpHeaderInfo("Content-Range", HttpHeaderType.Request | HttpHeaderType.Response) },
                { "ContentType", new HttpHeaderInfo("Content-Type", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.Restricted) },
                { "Cookie", new HttpHeaderInfo("Cookie", HttpHeaderType.Request) },
                { "Cookie2", new HttpHeaderInfo("Cookie2", HttpHeaderType.Request) },
                { "Date", new HttpHeaderInfo("Date", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.Restricted) },
                { "Expect", new HttpHeaderInfo("Expect", HttpHeaderType.Request | HttpHeaderType.Restricted | HttpHeaderType.MultiValue) },
                { "Expires", new HttpHeaderInfo("Expires", HttpHeaderType.Request | HttpHeaderType.Response) },
                { "ETag", new HttpHeaderInfo("ETag", HttpHeaderType.Response) },
                { "From", new HttpHeaderInfo("From", HttpHeaderType.Request) },
                { "Host", new HttpHeaderInfo("Host", HttpHeaderType.Request | HttpHeaderType.Restricted) },
                { "IfMatch", new HttpHeaderInfo("If-Match", HttpHeaderType.Request | HttpHeaderType.MultiValue) },
                { "IfModifiedSince", new HttpHeaderInfo("If-Modified-Since", HttpHeaderType.Request | HttpHeaderType.Restricted) },
                { "IfNoneMatch", new HttpHeaderInfo("If-None-Match", HttpHeaderType.Request | HttpHeaderType.MultiValue) },
                { "IfRange", new HttpHeaderInfo("If-Range", HttpHeaderType.Request) },
                { "IfUnmodifiedSince", new HttpHeaderInfo("If-Unmodified-Since", HttpHeaderType.Request) },
                { "KeepAlive", new HttpHeaderInfo("Keep-Alive", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.MultiValue) },
                { "LastModified", new HttpHeaderInfo("Last-Modified", HttpHeaderType.Request | HttpHeaderType.Response) },
                { "Location", new HttpHeaderInfo("Location", HttpHeaderType.Response) },
                { "MaxForwards", new HttpHeaderInfo("Max-Forwards", HttpHeaderType.Request) },
                { "Pragma", new HttpHeaderInfo("Pragma", HttpHeaderType.Request | HttpHeaderType.Response) },
                { "ProxyAuthenticate", new HttpHeaderInfo("Proxy-Authenticate", HttpHeaderType.Response | HttpHeaderType.MultiValue) },
                { "ProxyAuthorization", new HttpHeaderInfo("Proxy-Authorization", HttpHeaderType.Request) },
                { "ProxyConnection", new HttpHeaderInfo("Proxy-Connection", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.Restricted) },
                { "Public", new HttpHeaderInfo("Public", HttpHeaderType.Response | HttpHeaderType.MultiValue) },
                { "Range", new HttpHeaderInfo("Range", HttpHeaderType.Request | HttpHeaderType.Restricted | HttpHeaderType.MultiValue) },
                { "Referer", new HttpHeaderInfo("Referer", HttpHeaderType.Request | HttpHeaderType.Restricted) },
                { "RetryAfter", new HttpHeaderInfo("Retry-After", HttpHeaderType.Response) },
                { "SecWebSocketAccept", new HttpHeaderInfo("Sec-WebSocket-Accept", HttpHeaderType.Response | HttpHeaderType.Restricted) },
                { "SecWebSocketExtensions", new HttpHeaderInfo("Sec-WebSocket-Extensions", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.Restricted | HttpHeaderType.MultiValueInRequest) },
                { "SecWebSocketKey", new HttpHeaderInfo("Sec-WebSocket-Key", HttpHeaderType.Request | HttpHeaderType.Restricted) },
                { "SecWebSocketProtocol", new HttpHeaderInfo("Sec-WebSocket-Protocol", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.MultiValueInRequest) },
                { "SecWebSocketVersion", new HttpHeaderInfo("Sec-WebSocket-Version", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.Restricted | HttpHeaderType.MultiValueInResponse) },
                { "Server", new HttpHeaderInfo("Server", HttpHeaderType.Response) },
                { "SetCookie", new HttpHeaderInfo("Set-Cookie", HttpHeaderType.Response | HttpHeaderType.MultiValue) },
                { "SetCookie2", new HttpHeaderInfo("Set-Cookie2", HttpHeaderType.Response | HttpHeaderType.MultiValue) },
                { "Te", new HttpHeaderInfo("TE", HttpHeaderType.Request) },
                { "Trailer", new HttpHeaderInfo("Trailer", HttpHeaderType.Request | HttpHeaderType.Response) },
                { "TransferEncoding", new HttpHeaderInfo("Transfer-Encoding", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.Restricted | HttpHeaderType.MultiValue) },
                { "Translate", new HttpHeaderInfo("Translate", HttpHeaderType.Request) },
                { "Upgrade", new HttpHeaderInfo("Upgrade", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.MultiValue) },
                { "UserAgent", new HttpHeaderInfo("User-Agent", HttpHeaderType.Request | HttpHeaderType.Restricted) },
                { "Vary", new HttpHeaderInfo("Vary", HttpHeaderType.Response | HttpHeaderType.MultiValue) },
                { "Via", new HttpHeaderInfo("Via", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.MultiValue) },
                { "Warning", new HttpHeaderInfo("Warning", HttpHeaderType.Request | HttpHeaderType.Response | HttpHeaderType.MultiValue) },
                { "WwwAuthenticate", new HttpHeaderInfo("WWW-Authenticate", HttpHeaderType.Response | HttpHeaderType.Restricted | HttpHeaderType.MultiValue) }
            };
        }

        public WebHeaderCollection()
        {
        }

        /// <summary>
        /// Initializes a new instance of the WebHeaderCollection class with the specified parameters.
        /// </summary>
        /// <param name="state">The HTTP header type.</param>
        /// <param name="internallyUsed">A boolean indicating whether the collection is internally used.</param>
        internal WebHeaderCollection(HttpHeaderType state, bool internallyUsed)
        {
            _state = state;
            _internallyUsed = internallyUsed;
        }

        /// <summary>
        /// Initializes a new instance of the WebHeaderCollection class with serialization information.
        /// </summary>
        /// <param name="serializationInfo">The SerializationInfo containing the data needed to serialize the WebHeaderCollection.</param>
        /// <param name="streamingContext">The StreamingContext containing the source and destination of the serialized stream associated with the WebHeaderCollection.</param>
        protected WebHeaderCollection(
          SerializationInfo serializationInfo, StreamingContext streamingContext)
        {
            if (serializationInfo == null)
            {
                throw new ArgumentNullException(nameof(serializationInfo));
            }

            try
            {
                _internallyUsed = serializationInfo.GetBoolean("InternallyUsed");
                _state = (HttpHeaderType)serializationInfo.GetInt32("State");

                var cnt = serializationInfo.GetInt32("Count");
                for (var i = 0; i < cnt; i++)
                {
                    base.Add(
                      serializationInfo.GetString(i.ToString()),
                      serializationInfo.GetString((cnt + i).ToString()));
                }
            }
            catch (SerializationException ex)
            {
                throw new ArgumentException(ex.Message, nameof(serializationInfo), ex);
            }
        }

        public override string[] AllKeys => base.AllKeys;
        public override int Count => base.Count;
        public override KeysCollection Keys => base.Keys;
        internal HttpHeaderType State => _state;

        public string this[HttpRequestHeader header]
        {
            get
            {
                return Get(Convert(header));
            }

            set
            {
                Add(header, value);
            }
        }

        public string this[HttpResponseHeader header]
        {
            get
            {
                return Get(Convert(header));
            }

            set
            {
                Add(header, value);
            }
        }

        public static bool IsRestricted(string headerName)
        {
            return isRestricted(CheckName(headerName), false);
        }

        public static bool IsRestricted(string headerName, bool response)
        {
            return isRestricted(CheckName(headerName), response);
        }

        public void Add(string header)
        {
            if (header == null || header.Length == 0)
            {
                throw new ArgumentNullException(nameof(header));
            }

            var pos = checkColonSeparated(header);
            add(header.Substring(0, pos), header.Substring(pos + 1), false);
        }

        public void Add(HttpRequestHeader header, string value)
        {
            DoWithCheckingState(addWithoutCheckingName, Convert(header), value, false, true);
        }

        public void Add(HttpResponseHeader header, string value)
        {
            DoWithCheckingState(addWithoutCheckingName, Convert(header), value, true, true);
        }

        public override void Add(string name, string value)
        {
            add(name, value, false);
        }

        public override void Clear()
        {
            base.Clear();
            _state = HttpHeaderType.Unspecified;
        }

        public override string Get(int index)
        {
            return base.Get(index);
        }

        public override string Get(string name)
        {
            return base.Get(name);
        }

        public override IEnumerator GetEnumerator()
        {
            return base.GetEnumerator();
        }

        public override string GetKey(int index)
        {
            return base.GetKey(index);
        }

        [SecurityPermission(
          SecurityAction.LinkDemand, Flags = SecurityPermissionFlag.SerializationFormatter)]
        public override void GetObjectData(
          SerializationInfo serializationInfo, StreamingContext streamingContext)
        {
            if (serializationInfo == null)
            {
                throw new ArgumentNullException(nameof(serializationInfo));
            }

            serializationInfo.AddValue("InternallyUsed", _internallyUsed);
            serializationInfo.AddValue("State", (int)_state);

            var cnt = Count;
            serializationInfo.AddValue("Count", cnt);
            cnt.Times(
              i =>
              {
                  serializationInfo.AddValue(i.ToString(), GetKey(i));
                  serializationInfo.AddValue((cnt + i).ToString(), Get(i));
              });
        }

        [SecurityPermission(
          SecurityAction.LinkDemand,
          Flags = SecurityPermissionFlag.SerializationFormatter,
          SerializationFormatter = true)]
        void ISerializable.GetObjectData(
          SerializationInfo serializationInfo, StreamingContext streamingContext)
        {
            GetObjectData(serializationInfo, streamingContext);
        }

        public override string[] GetValues(int index)
        {
            var vals = base.GetValues(index);
            return vals != null && vals.Length > 0 ? vals : null;
        }

        public override string[] GetValues(string header)
        {
            var vals = base.GetValues(header);
            return vals != null && vals.Length > 0 ? vals : null;
        }

        public override void OnDeserialization(object sender)
        {
        }

        public void Remove(HttpRequestHeader header)
        {
            DoWithCheckingState(removeWithoutCheckingName, Convert(header), null, false, false);
        }

        public void Remove(HttpResponseHeader header)
        {
            DoWithCheckingState(removeWithoutCheckingName, Convert(header), null, true, false);
        }

        public override void Remove(string name)
        {
            DoWithCheckingState(removeWithoutCheckingName, CheckName(name), null, false);
        }

        public void Set(HttpRequestHeader header, string value)
        {
            DoWithCheckingState(setWithoutCheckingName, Convert(header), value, false, true);
        }

        public void Set(HttpResponseHeader header, string value)
        {
            DoWithCheckingState(setWithoutCheckingName, Convert(header), value, true, true);
        }

        public override void Set(string name, string value)
        {
            DoWithCheckingState(setWithoutCheckingName, CheckName(name), value, true);
        }

        /// <summary>
        /// Returns a byte array representing the WebHeaderCollection in UTF-8 encoding.
        /// </summary>
        /// <returns>A byte array representing the WebHeaderCollection.</returns>
        public byte[] ToByteArray()
        {
            return Encoding.UTF8.GetBytes(ToString());
        }

        /// <summary>
        /// Returns a string representation of the WebHeaderCollection.
        /// </summary>
        /// <returns>A string representing the WebHeaderCollection.</returns>
        public override string ToString()
        {
            var buff = new StringBuilder();
            Count.Times(i => buff.AppendFormat($"{GetKey(i)}: {Get(i)}\r\n"));

            return buff.Append("\r\n").ToString();
        }

        /// <summary>
        /// Converts the specified HttpRequestHeader to a string.
        /// </summary>
        /// <param name="header">The HttpRequestHeader to convert.</param>
        /// <returns>A string representing the converted HttpRequestHeader.</returns>
        internal static string Convert(HttpRequestHeader header)
        {
            return Convert(header.ToString());
        }

        internal static string Convert(HttpResponseHeader header)
        {
            return Convert(header.ToString());
        }

        internal static bool IsHeaderName(string name)
        {
            return name != null && name.Length > 0 && name.IsToken();
        }

        internal static bool IsHeaderValue(string value)
        {
            return value.IsText();
        }

        internal static bool IsMultiValue(string headerName, bool response)
        {
            if (headerName == null || headerName.Length == 0)
            {
                return false;
            }

            var info = GetHeaderInfo(headerName);
            return info != null && info.IsMultiValue(response);
        }

        internal void InternalRemove(string name)
        {
            base.Remove(name);
        }

        internal void InternalSet(string header, bool response)
        {
            var pos = checkColonSeparated(header);
            InternalSet(header.Substring(0, pos), header.Substring(pos + 1), response);
        }

        internal void InternalSet(string name, string value, bool response)
        {
            value = CheckValue(value);
            if (IsMultiValue(name, response))
            {
                base.Add(name, value);
            }
            else
            {
                base.Set(name, value);
            }
        }

        internal string ToStringMultiValue(bool response)
        {
            var buff = new StringBuilder();
            Count.Times(
              i =>
              {
                  var key = GetKey(i);
                  if (IsMultiValue(key, response))
                  {
                      foreach (var val in GetValues(i))
                      {
                          buff.AppendFormat($"{key}: {val}\r\n");
                      }
                  }
                  else
                  {
                      buff.AppendFormat($"{key}: {Get(i)}\r\n");
                  }
              });

            return buff.Append("\r\n").ToString();
        }

        protected void AddWithoutValidate(string headerName, string headerValue)
        {
            add(headerName, headerValue, true);
        }

        private static int checkColonSeparated(string header)
        {
            var idx = header.IndexOf(':');
            if (idx == -1)
            {
                throw new ArgumentException("No colon could be found.", nameof(header));
            }

            return idx;
        }

        private static HttpHeaderType CheckHeaderType(string name)
        {
            var info = GetHeaderInfo(name);
            return info == null
                   ? HttpHeaderType.Unspecified
                   : info.IsRequest && !info.IsResponse
                     ? HttpHeaderType.Request
                     : !info.IsRequest && info.IsResponse
                       ? HttpHeaderType.Response
                       : HttpHeaderType.Unspecified;
        }

        private static string CheckName(string name)
        {
            if (name == null || name.Length == 0)
            {
                throw new ArgumentNullException(nameof(name));
            }

            name = name.Trim();
            if (!IsHeaderName(name))
            {
                throw new ArgumentException("Contains invalid characters.", nameof(name));
            }

            return name;
        }

        private static string CheckValue(string value)
        {
            if (value == null || value.Length == 0)
            {
                return string.Empty;
            }

            value = value.Trim();
            if (value.Length > BUFFER_SIZE)
            {
                throw new ArgumentOutOfRangeException(nameof(value), "Greater than 65,535 characters.");
            }

            if (!IsHeaderValue(value))
            {
                throw new ArgumentException("Contains invalid characters.", nameof(value));
            }

            return value;
        }

        private static string Convert(string key)
        {
            return _headers.TryGetValue(key, out HttpHeaderInfo info) ? info.Name : string.Empty;
        }

        private static HttpHeaderInfo GetHeaderInfo(string name)
        {
            foreach (var info in _headers.Values)
            {
                if (info.Name.Equals(name, StringComparison.InvariantCultureIgnoreCase))
                {
                    return info;
                }
            }

            return null;
        }

        private static bool isRestricted(string name, bool response)
        {
            var info = GetHeaderInfo(name);
            return info != null && info.IsRestricted(response);
        }

        private void add(string name, string value, bool ignoreRestricted)
        {
            var act = ignoreRestricted
                      ? (Action<string, string>)addWithoutCheckingNameAndRestricted
                      : addWithoutCheckingName;

            DoWithCheckingState(act, CheckName(name), value, true);
        }

        private void addWithoutCheckingName(string name, string value)
        {
            DoWithoutCheckingName(base.Add, name, value);
        }

        private void addWithoutCheckingNameAndRestricted(string name, string value)
        {
            base.Add(name, CheckValue(value));
        }

        private void CheckRestricted(string name)
        {
            if (!_internallyUsed && isRestricted(name, true))
            {
                throw new ArgumentException("This header must be modified with the appropiate property.");
            }
        }

        private void CheckState(bool response)
        {
            if (_state == HttpHeaderType.Unspecified)
            {
                return;
            }

            if (response && _state == HttpHeaderType.Request)
            {
                throw new InvalidOperationException(
                  "This collection has already been used to store the request headers.");
            }

            if (!response && _state == HttpHeaderType.Response)
            {
                throw new InvalidOperationException(
                  "This collection has already been used to store the response headers.");
            }
        }

        private void DoWithCheckingState(
          Action<string, string> action, string name, string value, bool setState)
        {
            var type = CheckHeaderType(name);
            if (type == HttpHeaderType.Request)
            {
                DoWithCheckingState(action, name, value, false, setState);
            }
            else if (type == HttpHeaderType.Response)
            {
                DoWithCheckingState(action, name, value, true, setState);
            }
            else
            {
                action(name, value);
            }
        }

        private void DoWithCheckingState(
          Action<string, string> action, string name, string value, bool response, bool setState)
        {
            CheckState(response);
            action(name, value);
            if (setState && _state == HttpHeaderType.Unspecified)
            {
                _state = response ? HttpHeaderType.Response : HttpHeaderType.Request;
            }
        }

        private void DoWithoutCheckingName(Action<string, string> action, string name, string value)
        {
            CheckRestricted(name);
            action(name, CheckValue(value));
        }

        private void removeWithoutCheckingName(string name, string unuse)
        {
            CheckRestricted(name);
            base.Remove(name);
        }

        private void setWithoutCheckingName(string name, string value)
        {
            DoWithoutCheckingName(base.Set, name, value);
        }
    }
}