// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.IO;
using System.Net.Security;
using System.Net.Sockets;
using System.Security.Principal;
using System.Text;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents the context of a WebSocket connection within a <see cref="TcpListener"/>.
    /// </summary>
    /// <remarks>
    /// This internal class provides access to various properties and methods for interacting with the WebSocket connection
    /// within the context of a TCP listener.
    /// </remarks>
    internal class TcpListenerWSContext : WSContext
    {
        private readonly bool _secure;
        private readonly TcpClient _tcpClient;
        private readonly Uri _uri;
        private readonly WSClient _websocket;
        private CookieCollection _cookies;
        private NameValueCollection _queryString;
        private WebRequest _request;
        private IPrincipal _user;

        /// <summary>
        /// Initializes a new instance of the <see cref="TcpListenerWSContext"/> class.
        /// </summary>
        /// <param name="tcpClient">The <see cref="TcpClient"/> associated with the WebSocket connection.</param>
        /// <param name="protocol">The WebSocket protocol negotiated during the connection.</param>
        /// <param name="secure">A boolean indicating whether the connection is secure.</param>
        /// <param name="sslConfig">The SSL config for secure connections.</param>
        internal TcpListenerWSContext(
          TcpClient tcpClient,
          string protocol,
          bool secure,
          SSLConfigServer sslConfig
        )
        {
            _tcpClient = tcpClient;
            _secure = secure;

            var netStream = tcpClient.GetStream();
            if (secure)
            {
                var sslStream =
                  new SslStream(netStream, false, sslConfig.ClientCertificateValidationCallback);

                sslStream.AuthenticateAsServer(
                  sslConfig.Certificate,
                  sslConfig.IsClientCertificateRequired,
                  sslConfig.SslProtocols,
                  sslConfig.CheckForCertificateRevocation
                );

                Stream = sslStream;
            }
            else
            {
                Stream = netStream;
            }

            _request = WebRequest.Read(Stream, 90000);
            _uri =
              HttpUtility.CreateRequestUrl(
                _request.RequestUri, _request.Headers["Host"], _request.IsWebSocketRequest, secure
              );

            _websocket = new WSClient(this, protocol);
        }

        public override CookieCollection CookieCollection => _cookies ??= _request.Cookies;

        public override NameValueCollection Headers => _request.Headers;

        public override string Host => _request.Headers["Host"];

        public override bool IsAuthenticated => _user != null;

        public override bool IsLocal => UserEndPoint.Address.IsLocal();

        public override bool IsSecureConnection => _secure;

        public override bool IsWebSocketRequest => _request.IsWebSocketRequest;

        public override string Origin => _request.Headers["Origin"];

        public override NameValueCollection QueryString => _queryString ??=
                                   HttpUtility.InternalParseQueryString(
                                     _uri?.Query, Encoding.UTF8
                                   )
        ;

        public override Uri RequestUri => _uri;

        public override string SecWebSocketKey => _request.Headers["Sec-WebSocket-Key"];

        public override IEnumerable<string> SecWebSocketProtocols
        {
            get
            {
                var protocols = _request.Headers["Sec-WebSocket-Protocol"];
                if (protocols != null)
                {
                    foreach (var protocol in protocols.Split(','))
                    {
                        yield return protocol.Trim();
                    }
                }
            }
        }

        public override string SecWebSocketVersion => _request.Headers["Sec-WebSocket-Version"];

        public override System.Net.IPEndPoint ServerEndPoint => (System.Net.IPEndPoint)_tcpClient.Client.LocalEndPoint;

        public override IPrincipal User => _user;

        public override System.Net.IPEndPoint UserEndPoint => (System.Net.IPEndPoint)_tcpClient.Client.RemoteEndPoint;

        public override WSClient WebSocket => _websocket;

        /// <summary>
        /// Gets the stream of the underlying TCP connection.
        /// </summary>
        internal Stream Stream { get; }

        /// <inheritdoc />
        public override string ToString()
        {
            return _request.ToString();
        }

        /// <summary>
        /// Authenticates the WebSocket connection based on the specified authentication scheme.
        /// </summary>
        /// <param name="scheme">The authentication scheme to use.</param>
        /// <param name="realm">The authentication realm.</param>
        /// <param name="credentialsFinder">A function to find network credentials based on identity.</param>
        /// <returns>True if authentication is successful; otherwise, false.</returns>
        internal bool Authenticate(
          AuthenticationSchemes scheme,
          string realm,
          Func<IIdentity, NetworkCredential> credentialsFinder
        )
        {
            if (scheme == AuthenticationSchemes.Anonymous)
            {
                return true;
            }

            if (scheme == AuthenticationSchemes.None)
            {
                Close(HttpStatusCode.Forbidden);
                return false;
            }

            var chal = new AuthenticationChallenge(scheme, realm).ToString();

            var retry = -1;
            Func<bool> auth = null;
            auth =
              () =>
              {
                  retry++;
                  if (retry > 99)
                  {
                      Close(HttpStatusCode.Forbidden);
                      return false;
                  }

                  var user =
              HttpUtility.CreateUser(
                _request.Headers["Authorization"],
                scheme,
                realm,
                _request.HttpMethod,
                credentialsFinder
              );

                  if (user == null || !user.Identity.IsAuthenticated)
                  {
                      SendAuthenticationChallenge(chal);
                      return auth();
                  }

                  _user = user;
                  return true;
              };

            return auth();
        }

        /// <summary>
        /// Closes the WebSocket connection.
        /// </summary>
        internal void Close()
        {
            Stream.Close();
            _tcpClient.Close();
        }

        /// <summary>
        /// Closes the WebSocket connection with the specified HTTP status code.
        /// </summary>
        /// <param name="code">The HTTP status code indicating the reason for closure.</param>
        internal void Close(HttpStatusCode code)
        {
            _websocket.Close(WebResponse.CreateCloseResponse(code));
        }

        /// <summary>
        /// Sends an authentication challenge to the WebSocket client.
        /// </summary>
        /// <param name="challenge">The authentication challenge.</param>
        internal void SendAuthenticationChallenge(string challenge)
        {
            var buff = WebResponse.CreateUnauthorizedResponse(challenge).ToByteArray();
            Stream.Write(buff, 0, buff.Length);
            _request = WebRequest.Read(Stream, 15000);
        }
    }
}