// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Net;

namespace EonaCat.Network
{
    /// <summary>
    /// Manages HTTP endpoint listeners and their prefixes.
    /// </summary>
    internal sealed class EndPointManager
    {
        private static readonly Dictionary<IPEndPoint, EndPointListener> _endpoints;

        /// <summary>
        /// Initializes static members of the <see cref="EndPointManager"/> class.
        /// </summary>
        static EndPointManager()
        {
            _endpoints = new Dictionary<IPEndPoint, EndPointListener>();
        }

        /// <summary>
        /// Prevents a default instance of the <see cref="EndPointManager"/> class from being created.
        /// </summary>
        private EndPointManager()
        {
        }

        /// <summary>
        /// Adds an HTTP listener and its associated prefixes.
        /// </summary>
        /// <param name="listener">The HTTP listener to be added.</param>
        public static void AddListener(HttpListener listener)
        {
            var added = new List<string>();
            lock (((ICollection)_endpoints).SyncRoot)
            {
                try
                {
                    foreach (var pref in listener.Prefixes)
                    {
                        addPrefix(pref, listener);
                        added.Add(pref);
                    }
                }
                catch
                {
                    foreach (var pref in added)
                    {
                        removePrefix(pref, listener);
                    }

                    throw;
                }
            }
        }

        /// <summary>
        /// Adds an HTTP listener prefix.
        /// </summary>
        /// <param name="uriPrefix">The URI prefix to be added.</param>
        /// <param name="listener">The HTTP listener associated with the prefix.</param>
        public static void AddPrefix(string uriPrefix, HttpListener listener)
        {
            lock (((ICollection)_endpoints).SyncRoot)
            {
                addPrefix(uriPrefix, listener);
            }
        }

        /// <summary>
        /// Removes an HTTP listener and its associated prefixes.
        /// </summary>
        /// <param name="listener">The HTTP listener to be removed.</param>
        public static void RemoveListener(HttpListener listener)
        {
            lock (((ICollection)_endpoints).SyncRoot)
            {
                foreach (var pref in listener.Prefixes)
                {
                    removePrefix(pref, listener);
                }
            }
        }

        /// <summary>
        /// Removes an HTTP listener prefix.
        /// </summary>
        /// <param name="uriPrefix">The URI prefix to be removed.</param>
        /// <param name="listener">The HTTP listener associated with the prefix.</param>
        public static void RemovePrefix(string uriPrefix, HttpListener listener)
        {
            lock (((ICollection)_endpoints).SyncRoot)
            {
                removePrefix(uriPrefix, listener);
            }
        }

        /// <summary>
        /// Removes an endpoint and closes its associated listener.
        /// </summary>
        /// <param name="endpoint">The endpoint to be removed.</param>
        /// <returns><c>true</c> if the endpoint is successfully removed; otherwise, <c>false</c>.</returns>
        internal static bool RemoveEndPoint(IPEndPoint endpoint)
        {
            lock (((ICollection)_endpoints).SyncRoot)
            {
                if (!_endpoints.TryGetValue(endpoint, out EndPointListener lsnr))
                {
                    return false;
                }

                _endpoints.Remove(endpoint);
                lsnr.Close();

                return true;
            }
        }

        private static void addPrefix(string uriPrefix, HttpListener listener)
        {
            var pref = new HttpListenerPrefix(uriPrefix);

            var addr = convertToIPAddress(pref.Host);
            if (!addr.IsLocal())
            {
                throw new HttpListenerException(87, "Includes an invalid host.");
            }

            if (!int.TryParse(pref.Port, out int port))
            {
                throw new HttpListenerException(87, "Includes an invalid port.");
            }

            if (!port.IsPortNumber())
            {
                throw new HttpListenerException(87, "Includes an invalid port.");
            }

            var path = pref.Path;
            if (path.IndexOf('%') != -1)
            {
                throw new HttpListenerException(87, "Includes an invalid path.");
            }

            if (path.IndexOf("//", StringComparison.Ordinal) != -1)
            {
                throw new HttpListenerException(87, "Includes an invalid path.");
            }

            var endpoint = new IPEndPoint(addr, port);

            if (_endpoints.TryGetValue(endpoint, out EndPointListener lsnr))
            {
                if (lsnr.IsSecure ^ pref.IsSecure)
                {
                    throw new HttpListenerException(87, "Includes an invalid scheme.");
                }
            }
            else
            {
                lsnr =
                  new EndPointListener(
                    endpoint,
                    pref.IsSecure,
                    listener.CertificateFolderPath,
                    listener.SSL,
                    listener.ReuseAddress
                  );

                _endpoints.Add(endpoint, lsnr);
            }

            lsnr.AddPrefix(pref, listener);
        }

        private static IPAddress convertToIPAddress(string hostname)
        {
            return hostname == "*" || hostname == "+" ? IPAddress.Any : hostname.ToIPAddress();
        }

        private static void removePrefix(string uriPrefix, HttpListener listener)
        {
            var pref = new HttpListenerPrefix(uriPrefix);

            var addr = convertToIPAddress(pref.Host);
            if (!addr.IsLocal())
            {
                return;
            }

            if (!int.TryParse(pref.Port, out int port))
            {
                return;
            }

            if (!port.IsPortNumber())
            {
                return;
            }

            var path = pref.Path;
            if (path.IndexOf('%') != -1)
            {
                return;
            }

            if (path.IndexOf("//", StringComparison.Ordinal) != -1)
            {
                return;
            }

            var endpoint = new IPEndPoint(addr, port);

            if (!_endpoints.TryGetValue(endpoint, out EndPointListener lsnr))
            {
                return;
            }

            if (lsnr.IsSecure ^ pref.IsSecure)
            {
                return;
            }

            lsnr.RemovePrefix(pref, listener);
        }
    }
}