// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

namespace EonaCat.Network
{
    /// <summary>
    /// Represents information about an HTTP header.
    /// </summary>
    internal class HttpHeaderInfo
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="HttpHeaderInfo"/> class.
        /// </summary>
        /// <param name="name">The name of the header.</param>
        /// <param name="type">The type of the header.</param>
        internal HttpHeaderInfo(string name, HttpHeaderType type)
        {
            Name = name;
            Type = type;
        }

        /// <summary>
        /// Gets a value indicating whether the header is for a request.
        /// </summary>
        public bool IsRequest => (Type & HttpHeaderType.Request) == HttpHeaderType.Request;

        /// <summary>
        /// Gets a value indicating whether the header is for a response.
        /// </summary>
        public bool IsResponse => (Type & HttpHeaderType.Response) == HttpHeaderType.Response;

        /// <summary>
        /// Gets the name of the header.
        /// </summary>
        public string Name { get; }

        /// <summary>
        /// Gets the type of the header.
        /// </summary>
        public HttpHeaderType Type { get; }

        /// <summary>
        /// Gets a value indicating whether the header is multi-value in a request.
        /// </summary>
        internal bool IsMultiValueInRequest => (Type & HttpHeaderType.MultiValueInRequest) == HttpHeaderType.MultiValueInRequest;

        /// <summary>
        /// Gets a value indicating whether the header is multi-value in a response.
        /// </summary>
        internal bool IsMultiValueInResponse => (Type & HttpHeaderType.MultiValueInResponse) == HttpHeaderType.MultiValueInResponse;

        /// <summary>
        /// Gets a value indicating whether the header is multi-value.
        /// </summary>
        /// <param name="response">True if checking for a response; false for a request.</param>
        /// <returns>True if the header is multi-value, false otherwise.</returns>
        public bool IsMultiValue(bool response)
        {
            return (Type & HttpHeaderType.MultiValue) == HttpHeaderType.MultiValue
                   ? (response ? IsResponse : IsRequest)
                   : (response ? IsMultiValueInResponse : IsMultiValueInRequest);
        }

        /// <summary>
        /// Gets a value indicating whether the header is restricted.
        /// </summary>
        /// <param name="response">True if checking for a response; false for a request.</param>
        /// <returns>True if the header is restricted, false otherwise.</returns>
        public bool IsRestricted(bool response)
        {
            return (Type & HttpHeaderType.Restricted) == HttpHeaderType.Restricted && (response ? IsResponse : IsRequest);
        }
    }
}