// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Threading;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents the result of an asynchronous operation for an <see cref="HttpListener"/>.
    /// </summary>
    internal class HttpListenerAsyncResult : IAsyncResult
    {
        private readonly AsyncCallback _callback;
        private readonly object _locker;
        private bool _completed;
        private HttpListenerContext _context;
        private Exception _exception;
        private ManualResetEvent _waitHandle;

        /// <summary>
        /// Initializes a new instance of the <see cref="HttpListenerAsyncResult"/> class.
        /// </summary>
        /// <param name="callback">The method to call when the operation completes.</param>
        /// <param name="state">A user-defined object that contains information about the asynchronous operation.</param>
        internal HttpListenerAsyncResult(AsyncCallback callback, object state)
        {
            _callback = callback;
            AsyncState = state;
            _locker = new object();
        }

        /// <summary>
        /// Gets the user-defined object that contains information about the asynchronous operation.
        /// </summary>
        public object AsyncState { get; }

        /// <summary>
        /// Gets a <see cref="WaitHandle"/> that is used to wait for an asynchronous operation to complete.
        /// </summary>
        public WaitHandle AsyncWaitHandle
        {
            get
            {
                lock (_locker)
                {
                    return _waitHandle ??= new ManualResetEvent(_completed);
                }
            }
        }

        /// <summary>
        /// Gets a value indicating whether the asynchronous operation completed synchronously.
        /// </summary>
        public bool CompletedSynchronously { get; private set; }

        /// <summary>
        /// Gets a value indicating whether the asynchronous operation has completed.
        /// </summary>
        public bool IsCompleted
        {
            get
            {
                lock (_locker)
                {
                    return _completed;
                }
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the <see cref="EndGetContext"/> method has been called.
        /// </summary>
        internal bool EndCalled { get; set; }

        /// <summary>
        /// Gets or sets a value indicating whether the asynchronous operation is in progress.
        /// </summary>
        internal bool InGet { get; set; }

        /// <summary>
        /// Completes the asynchronous operation with the specified exception.
        /// </summary>
        /// <param name="exception">The exception that occurred during the asynchronous operation.</param>
        internal void Complete(Exception exception)
        {
            _exception = InGet && (exception is ObjectDisposedException)
                         ? new HttpListenerException(995, "The listener is closed.")
                         : exception;

            complete(this);
        }

        /// <summary>
        /// Completes the asynchronous operation with the specified <see cref="HttpListenerContext"/>.
        /// </summary>
        /// <param name="context">The <see cref="HttpListenerContext"/> representing the result of the asynchronous operation.</param>
        internal void Complete(HttpListenerContext context)
        {
            Complete(context, false);
        }

        /// <summary>
        /// Completes the asynchronous operation with the specified <see cref="HttpListenerContext"/>.
        /// </summary>
        /// <param name="context">The <see cref="HttpListenerContext"/> representing the result of the asynchronous operation.</param>
        /// <param name="syncCompleted">A value indicating whether the completion was synchronous.</param>
        internal void Complete(HttpListenerContext context, bool syncCompleted)
        {
            _context = context;
            CompletedSynchronously = syncCompleted;

            complete(this);
        }

        /// <summary>
        /// Gets the <see cref="HttpListenerContext"/> result of the asynchronous operation.
        /// </summary>
        /// <returns>The <see cref="HttpListenerContext"/> representing the result of the asynchronous operation.</returns>
        internal HttpListenerContext GetContext()
        {
            if (_exception != null)
            {
                throw _exception;
            }

            return _context;
        }

        // Private method to complete the asynchronous operation
        private static void complete(HttpListenerAsyncResult asyncResult)
        {
            lock (asyncResult._locker)
            {
                asyncResult._completed = true;

                var waitHandle = asyncResult._waitHandle;
                waitHandle?.Set();
            }

            var callback = asyncResult._callback;
            if (callback == null)
            {
                return;
            }

            ThreadPool.QueueUserWorkItem(
              state =>
              {
                  try
                  {
                      callback(asyncResult);
                  }
                  catch
                  {
                  }
              },
              null
            );
        }
    }
}