// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents a URI prefix for an <see cref="HttpListener"/>.
    /// </summary>
    internal sealed class HttpListenerPrefix
    {
        private string _prefix;

        /// <summary>
        /// Initializes a new instance of the <see cref="HttpListenerPrefix"/> class with the specified URI prefix.
        /// </summary>
        /// <param name="uriPrefix">The URI prefix.</param>
        internal HttpListenerPrefix(string uriPrefix)
        {
            Original = uriPrefix;
            parse(uriPrefix);
        }

        /// <summary>
        /// Gets the host specified in the URI prefix.
        /// </summary>
        public string Host { get; private set; }

        /// <summary>
        /// Gets a value indicating whether the URI prefix is for a secure connection (https).
        /// </summary>
        public bool IsSecure { get; private set; }

        /// <summary>
        /// Gets or sets the associated <see cref="HttpListener"/>.
        /// </summary>
        public HttpListener Listener { get; set; }

        /// <summary>
        /// Gets the original URI prefix.
        /// </summary>
        public string Original { get; }

        /// <summary>
        /// Gets the path specified in the URI prefix.
        /// </summary>
        public string Path { get; private set; }

        /// <summary>
        /// Gets the port specified in the URI prefix.
        /// </summary>
        public string Port { get; private set; }

        /// <summary>
        /// Checks if the specified URI prefix is valid.
        /// </summary>
        /// <param name="uriPrefix">The URI prefix to check.</param>
        public static void CheckPrefix(string uriPrefix)
        {
            if (uriPrefix == null)
            {
                throw new ArgumentNullException(nameof(uriPrefix));
            }

            var len = uriPrefix.Length;
            if (len == 0)
            {
                throw new ArgumentException("An empty string.", nameof(uriPrefix));
            }

            if (!(uriPrefix.StartsWith("http://") || uriPrefix.StartsWith("https://")))
            {
                throw new ArgumentException("The scheme isn't 'http' or 'https'.", nameof(uriPrefix));
            }

            var startHost = uriPrefix.IndexOf(':') + 3;
            if (startHost >= len)
            {
                throw new ArgumentException("No host is specified.", nameof(uriPrefix));
            }

            if (uriPrefix[startHost] == ':')
            {
                throw new ArgumentException("No host is specified.", nameof(uriPrefix));
            }

            var root = uriPrefix.IndexOf('/', startHost, len - startHost);
            if (root == startHost)
            {
                throw new ArgumentException("No host is specified.", nameof(uriPrefix));
            }

            if (root == -1 || uriPrefix[len - 1] != '/')
            {
                throw new ArgumentException("Ends without '/'.", nameof(uriPrefix));
            }

            if (uriPrefix[root - 1] == ':')
            {
                throw new ArgumentException("No port is specified.", nameof(uriPrefix));
            }

            if (root == len - 2)
            {
                throw new ArgumentException("No path is specified.", nameof(uriPrefix));
            }
        }

        /// <summary>
        /// Determines whether the specified object is equal to the current object.
        /// </summary>
        /// <param name="obj">The object to compare with the current object.</param>
        /// <returns>true if the specified object is equal to the current object; otherwise, false.</returns>
        public override bool Equals(object obj)
        {
            return obj is HttpListenerPrefix pref && pref._prefix == _prefix;
        }

        /// <summary>
        /// Serves as the default hash function.
        /// </summary>
        /// <returns>A hash code for the current object.</returns>
        public override int GetHashCode()
        {
            return _prefix.GetHashCode();
        }

        /// <summary>
        /// Returns a string that represents the current object.
        /// </summary>
        /// <returns>A string that represents the current object.</returns>
        public override string ToString()
        {
            return _prefix;
        }

        private void parse(string uriPrefix)
        {
            if (uriPrefix.StartsWith("https"))
            {
                IsSecure = true;
            }

            var len = uriPrefix.Length;
            var startHost = uriPrefix.IndexOf(':') + 3;
            var root = uriPrefix.IndexOf('/', startHost + 1, len - startHost - 1);

            var colon = uriPrefix.LastIndexOf(':', root - 1, root - startHost - 1);
            if (uriPrefix[root - 1] != ']' && colon > startHost)
            {
                Host = uriPrefix.Substring(startHost, colon - startHost);
                Port = uriPrefix.Substring(colon + 1, root - colon - 1);
            }
            else
            {
                Host = uriPrefix.Substring(startHost, root - startHost);
                Port = IsSecure ? "443" : "80";
            }

            Path = uriPrefix.Substring(root);

            _prefix =
              string.Format("http{0}://{1}:{2}{3}", IsSecure ? "s" : "", Host, Port, Path);
        }
    }
}