// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections;
using System.Collections.Generic;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents a collection of URI prefixes for an <see cref="HttpListener"/>.
    /// </summary>
    public class HttpListenerPrefixCollection : ICollection<string>, IEnumerable<string>, IEnumerable
    {
        private readonly HttpListener _listener;
        private readonly List<string> _prefixes;

        /// <summary>
        /// Initializes a new instance of the <see cref="HttpListenerPrefixCollection"/> class.
        /// </summary>
        /// <param name="listener">The associated <see cref="HttpListener"/>.</param>
        internal HttpListenerPrefixCollection(HttpListener listener)
        {
            _listener = listener;
            _prefixes = new List<string>();
        }

        /// <summary>
        /// Gets the number of URI prefixes in the collection.
        /// </summary>
        public int Count => _prefixes.Count;

        /// <summary>
        /// Gets a value indicating whether the collection is read-only.
        /// </summary>
        public bool IsReadOnly => false;

        /// <summary>
        /// Gets a value indicating whether access to the collection is synchronized (thread-safe).
        /// </summary>
        public bool IsSynchronized => false;

        /// <summary>
        /// Adds a URI prefix to the collection.
        /// </summary>
        /// <param name="uriPrefix">The URI prefix to add.</param>
        public void Add(string uriPrefix)
        {
            _listener.CheckDisposed();
            HttpListenerPrefix.CheckPrefix(uriPrefix);
            if (_prefixes.Contains(uriPrefix))
            {
                return;
            }

            _prefixes.Add(uriPrefix);
            if (_listener.IsListening)
            {
                EndPointManager.AddPrefix(uriPrefix, _listener);
            }
        }

        /// <summary>
        /// Removes all URI prefixes from the collection.
        /// </summary>
        public void Clear()
        {
            _listener.CheckDisposed();
            _prefixes.Clear();
            if (_listener.IsListening)
            {
                EndPointManager.RemoveListener(_listener);
            }
        }

        /// <summary>
        /// Determines whether the collection contains a specific URI prefix.
        /// </summary>
        /// <param name="uriPrefix">The URI prefix to locate.</param>
        /// <returns>true if the collection contains the specified URI prefix; otherwise, false.</returns>
        public bool Contains(string uriPrefix)
        {
            _listener.CheckDisposed();
            if (uriPrefix == null)
            {
                throw new ArgumentNullException(nameof(uriPrefix));
            }

            return _prefixes.Contains(uriPrefix);
        }

        /// <summary>
        /// Copies the URI prefixes to an array, starting at a particular array index.
        /// </summary>
        /// <param name="array">The one-dimensional array that is the destination of the elements copied from the collection.</param>
        /// <param name="offset">The zero-based index in the array at which copying begins.</param>
        public void CopyTo(Array array, int offset)
        {
            _listener.CheckDisposed();
            ((ICollection)_prefixes).CopyTo(array, offset);
        }

        /// <summary>
        /// Copies the URI prefixes to an array, starting at a particular array index.
        /// </summary>
        /// <param name="array">The one-dimensional array that is the destination of the elements copied from the collection.</param>
        /// <param name="offset">The zero-based index in the array at which copying begins.</param>
        public void CopyTo(string[] array, int offset)
        {
            _listener.CheckDisposed();
            _prefixes.CopyTo(array, offset);
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>An enumerator that can be used to iterate through the collection.</returns>
        public IEnumerator<string> GetEnumerator()
        {
            return _prefixes.GetEnumerator();
        }

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <returns>An enumerator that can be used to iterate through the collection.</returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return _prefixes.GetEnumerator();
        }

        /// <summary>
        /// Removes the specified URI prefix from the collection.
        /// </summary>
        /// <param name="uriPrefix">The URI prefix to remove.</param>
        /// <returns>true if the URI prefix is successfully removed; otherwise, false. This method also returns false if the URI prefix was not found in the collection.</returns>
        public bool Remove(string uriPrefix)
        {
            _listener.CheckDisposed();
            if (uriPrefix == null)
            {
                throw new ArgumentNullException(nameof(uriPrefix));
            }

            var ret = _prefixes.Remove(uriPrefix);
            if (ret && _listener.IsListening)
            {
                EndPointManager.RemovePrefix(uriPrefix, _listener);
            }

            return ret;
        }
    }
}