// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Text;

namespace EonaCat.Network
{
    /// <summary>
    /// Represents an HTTP listener response.
    /// </summary>
    public sealed class HttpListenerResponse : IDisposable
    {
        private readonly HttpListenerContext _context;
        private Encoding _contentEncoding;
        private long _contentLength;
        private string _contentType;
        private CookieCollection _cookies;
        private bool _disposed;
        private WebHeaderCollection _headers;
        private bool _keepAlive;
        private string _location;
        private ResponseStream _outputStream;
        private bool _sendInChunks;
        private int _statusCode;
        private string _statusDescription;
        private Version _version;

        /// <summary>
        /// Initializes a new instance of the <see cref="HttpListenerResponse"/> class.
        /// </summary>
        /// <param name="context">The HTTP listener context associated with the response.</param>
        internal HttpListenerResponse(HttpListenerContext context)
        {
            _context = context;
            _keepAlive = true;
            _statusCode = 200;
            _statusDescription = "OK";
            _version = HttpVersion.Version11;
        }

        /// <summary>
        /// Gets or sets the encoding for the response content.
        /// </summary>
        public Encoding ContentEncoding
        {
            get
            {
                return _contentEncoding;
            }

            set
            {
                checkDisposed();
                _contentEncoding = value;
            }
        }

        /// <summary>
        /// Gets or sets the content length of the response content.
        /// </summary>
        public long ContentLength64
        {
            get
            {
                return _contentLength;
            }

            set
            {
                checkDisposedOrHeadersSent();
                if (value < 0)
                {
                    throw new ArgumentOutOfRangeException("Less than zero.", "value");
                }

                _contentLength = value;
            }
        }

        /// <summary>
        /// Gets or sets the content type of the response content.
        /// </summary>
        public string ContentType
        {
            get
            {
                return _contentType;
            }

            set
            {
                checkDisposed();
                if (value != null && value.Length == 0)
                {
                    throw new ArgumentException("An empty string.", nameof(value));
                }

                _contentType = value;
            }
        }

        /// <summary>
        /// Gets or sets the collection of cookies in the response.
        /// </summary>
        public CookieCollection Cookies
        {
            get
            {
                return _cookies ??= new CookieCollection();
            }

            set
            {
                _cookies = value;
            }
        }

        /// <summary>
        /// Gets or sets the collection of headers in the response.
        /// </summary>
        public WebHeaderCollection Headers
        {
            get
            {
                return _headers ??= new WebHeaderCollection(HttpHeaderType.Response, false);
            }

            set
            {
                if (value != null && value.State != HttpHeaderType.Response)
                {
                    throw new InvalidOperationException(
                      "The specified headers aren't valid for a response.");
                }

                _headers = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the connection should be kept alive.
        /// </summary>
        public bool KeepAlive
        {
            get
            {
                return _keepAlive;
            }

            set
            {
                checkDisposedOrHeadersSent();
                _keepAlive = value;
            }
        }

        /// <summary>
        /// Gets the output stream for writing the response content.
        /// </summary>
        public Stream OutputStream
        {
            get
            {
                checkDisposed();
                return _outputStream ??= _context.Connection.GetResponseStream();
            }
        }

        /// <summary>
        /// Gets or sets the HTTP protocol version of the response.
        /// </summary>
        public Version ProtocolVersion
        {
            get
            {
                return _version;
            }

            set
            {
                checkDisposedOrHeadersSent();
                if (value == null)
                {
                    throw new ArgumentNullException(nameof(value));
                }

                if (value.Major != 1 || (value.Minor != 0 && value.Minor != 1))
                {
                    throw new ArgumentException("Not 1.0 or 1.1.", nameof(value));
                }

                _version = value;
            }
        }

        /// <summary>
        /// Gets or sets the redirect location for the response.
        /// </summary>
        public string RedirectLocation
        {
            get
            {
                return _location;
            }

            set
            {
                checkDisposed();
                if (value == null)
                {
                    _location = null;
                    return;
                }

                if (!value.MaybeUri() || !Uri.TryCreate(value, UriKind.Absolute, out Uri uri))
                {
                    throw new ArgumentException("Not an absolute URL.", nameof(value));
                }

                _location = value;
            }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the response should be sent in chunks.
        /// </summary>
        public bool SendInChunks
        {
            get
            {
                return _sendInChunks;
            }

            set
            {
                checkDisposedOrHeadersSent();
                _sendInChunks = value;
            }
        }

        /// <summary>
        /// Gets or sets the HTTP status code for the response.
        /// </summary>
        public int StatusCode
        {
            get
            {
                return _statusCode;
            }

            set
            {
                checkDisposedOrHeadersSent();
                if (value < 100 || value > 999)
                {
                    throw new System.Net.ProtocolViolationException(
                      "A value isn't between 100 and 999 inclusive.");
                }

                _statusCode = value;
                _statusDescription = value.GetStatusDescription();
            }
        }

        /// <summary>
        /// Gets or sets the status description for the response.
        /// </summary>
        public string StatusDescription
        {
            get
            {
                return _statusDescription;
            }

            set
            {
                checkDisposedOrHeadersSent();
                if (value == null || value.Length == 0)
                {
                    _statusDescription = _statusCode.GetStatusDescription();
                    return;
                }

                if (!value.IsText() || value.IndexOfAny(new[] { '\r', '\n' }) > -1)
                {
                    throw new ArgumentException("Contains invalid characters.", nameof(value));
                }

                _statusDescription = value;
            }
        }

        internal bool CloseConnection { get; set; }

        internal bool HeadersSent { get; set; }

        /// <summary>
        /// Aborts the response.
        /// </summary>
        public void Abort()
        {
            if (_disposed)
            {
                return;
            }

            close(true);
        }

        /// <summary>
        /// Appends a header to the response.
        /// </summary>
        /// <param name="name">The name of the header.</param>
        /// <param name="value">The value of the header.</param>
        public void AddHeader(string name, string value)
        {
            Headers.Set(name, value);
        }

        /// <summary>
        /// Appends a cookie to the response.
        /// </summary>
        /// <param name="cookie">The cookie to append.</param>
        public void AppendCookie(Cookie cookie)
        {
            Cookies.Add(cookie);
        }

        /// <summary>
        /// Appends a header to the response.
        /// </summary>
        /// <param name="name">The name of the header.</param>
        /// <param name="value">The value of the header.</param>
        public void AppendHeader(string name, string value)
        {
            Headers.Add(name, value);
        }

        /// <summary>
        /// Closes the response.
        /// </summary>
        public void Close()
        {
            if (_disposed)
            {
                return;
            }

            close(false);
        }

        /// <summary>
        /// Closes the response with the specified response entity and blocking behavior.
        /// </summary>
        /// <param name="responseEntity">The response entity.</param>
        /// <param name="willBlock">A value indicating whether the operation will block.</param>
        public void Close(byte[] responseEntity, bool willBlock)
        {
            checkDisposed();
            if (responseEntity == null)
            {
                throw new ArgumentNullException(nameof(responseEntity));
            }

            var len = responseEntity.Length;
            var output = OutputStream;
            if (willBlock)
            {
                output.Write(responseEntity, 0, len);
                close(false);

                return;
            }

            output.BeginWrite(
              responseEntity,
              0,
              len,
              ar =>
              {
                  output.EndWrite(ar);
                  close(false);
              },
              null);
        }

        /// <summary>
        /// Copies headers and other properties from a template response.
        /// </summary>
        /// <param name="templateResponse">The template response to copy from.</param>
        public void CopyFrom(HttpListenerResponse templateResponse)
        {
            if (templateResponse == null)
            {
                throw new ArgumentNullException(nameof(templateResponse));
            }

            if (templateResponse._headers != null)
            {
                _headers?.Clear();

                Headers.Add(templateResponse._headers);
            }
            else if (_headers != null)
            {
                _headers = null;
            }

            _contentLength = templateResponse._contentLength;
            _statusCode = templateResponse._statusCode;
            _statusDescription = templateResponse._statusDescription;
            _keepAlive = templateResponse._keepAlive;
            _version = templateResponse._version;
        }

        /// <inheritdoc/>
        void IDisposable.Dispose()
        {
            if (_disposed)
            {
                return;
            }

            close(true);
        }

        /// <summary>
        /// Redirects the response to the specified URL.
        /// </summary>
        /// <param name="url">The URL to redirect to.</param>
        public void Redirect(string url)
        {
            checkDisposedOrHeadersSent();
            if (url == null)
            {
                throw new ArgumentNullException(nameof(url));
            }

            if (!url.MaybeUri() || !Uri.TryCreate(url, UriKind.Absolute, out Uri uri))
            {
                throw new ArgumentException("Not an absolute URL.", nameof(url));
            }

            _location = url;
            _statusCode = 302;
            _statusDescription = "Found";
        }

        /// <summary>
        /// Sets a cookie in the response.
        /// </summary>
        /// <param name="cookie">The cookie to set.</param>
        public void SetCookie(Cookie cookie)
        {
            if (cookie == null)
            {
                throw new ArgumentNullException(nameof(cookie));
            }

            if (!canAddOrUpdate(cookie))
            {
                throw new ArgumentException("Cannot be replaced.", nameof(cookie));
            }

            Cookies.Add(cookie);
        }

        /// <summary>
        /// Writes headers to the specified destination.
        /// </summary>
        /// <param name="destination">The destination stream to write headers to.</param>
        /// <returns>The collection of headers written.</returns>
        internal WebHeaderCollection WriteHeadersTo(MemoryStream destination)
        {
            var headers = new WebHeaderCollection(HttpHeaderType.Response, true);
            if (_headers != null)
            {
                headers.Add(_headers);
            }

            if (_contentType != null)
            {
                var type = _contentType.IndexOf("charset=", StringComparison.Ordinal) == -1 &&
                           _contentEncoding != null
                           ? string.Format("{0}; charset={1}", _contentType, _contentEncoding.WebName)
                           : _contentType;

                headers.InternalSet("Content-Type", type, true);
            }

            if (headers["Server"] == null)
            {
                headers.InternalSet("Server", $"EonaCat.Network/{Constants.Version}", true);
            }

            var prov = CultureInfo.InvariantCulture;
            if (headers["Date"] == null)
            {
                headers.InternalSet("Date", DateTime.UtcNow.ToString("r", prov), true);
            }

            if (!_sendInChunks)
            {
                headers.InternalSet("Content-Length", _contentLength.ToString(prov), true);
            }
            else
            {
                headers.InternalSet("Transfer-Encoding", "chunked", true);
            }

            /*
             * Apache forces closing the connection for these status codes:
             * - 400 Bad Request
             * - 408 Request Timeout
             * - 411 Length Required
             * - 413 Request Entity Too Large
             * - 414 Request-Uri Too Long
             * - 500 Internal Server Error
             * - 503 Service Unavailable
             */
            var closeConn = !_context.Request.KeepAlive ||
                            !_keepAlive ||
                            _statusCode == 400 ||
                            _statusCode == 408 ||
                            _statusCode == 411 ||
                            _statusCode == 413 ||
                            _statusCode == 414 ||
                            _statusCode == 500 ||
                            _statusCode == 503;

            var reuses = _context.Connection.Reuses;
            if (closeConn || reuses >= 100)
            {
                headers.InternalSet("Connection", "close", true);
            }
            else
            {
                headers.InternalSet(
                  "Keep-Alive", string.Format("timeout=15,max={0}", 100 - reuses), true);

                if (_context.Request.ProtocolVersion < HttpVersion.Version11)
                {
                    headers.InternalSet("Connection", "keep-alive", true);
                }
            }

            if (_location != null)
            {
                headers.InternalSet("Location", _location, true);
            }

            if (_cookies != null)
            {
                foreach (Cookie cookie in _cookies)
                {
                    headers.InternalSet("Set-Cookie", cookie.ToResponseString(), true);
                }
            }

            var enc = _contentEncoding ?? Encoding.Default;
            var writer = new StreamWriter(destination, enc, 256);
            writer.Write("HTTP/{0} {1} {2}\r\n", _version, _statusCode, _statusDescription);
            writer.Write(headers.ToStringMultiValue(true));
            writer.Flush();

            // Assumes that the destination was at position 0.
            destination.Position = enc.GetPreamble().Length;

            return headers;
        }

        private bool canAddOrUpdate(Cookie cookie)
        {
            if (_cookies == null || _cookies.Count == 0)
            {
                return true;
            }

            var found = findCookie(cookie).ToList();
            if (found.Count == 0)
            {
                return true;
            }

            var ver = cookie.Version;
            foreach (var c in found)
            {
                if (c.Version == ver)
                {
                    return true;
                }
            }

            return false;
        }

        private void checkDisposed()
        {
            if (_disposed)
            {
                throw new ObjectDisposedException(GetType().ToString());
            }
        }

        private void checkDisposedOrHeadersSent()
        {
            if (_disposed)
            {
                throw new ObjectDisposedException(GetType().ToString());
            }

            if (HeadersSent)
            {
                throw new InvalidOperationException("Cannot be changed after the headers are sent.");
            }
        }

        private void close(bool force)
        {
            _disposed = true;
            _context.Connection.Close(force);
        }

        private IEnumerable<Cookie> findCookie(Cookie cookie)
        {
            var name = cookie.Name;
            var domain = cookie.Domain;
            var path = cookie.Path;
            if (_cookies != null)
            {
                foreach (Cookie c in _cookies)
                {
                    if (c.Name.Equals(name, StringComparison.OrdinalIgnoreCase) &&
                        c.Domain.Equals(domain, StringComparison.OrdinalIgnoreCase) &&
                        c.Path.Equals(path, StringComparison.Ordinal))
                    {
                        yield return c;
                    }
                }
            }
        }
    }
}