// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Net.Security;
using System.Net.Sockets;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace EonaCat.Network
{
    public class WSClient : IDisposable
    {
        internal static readonly byte[] EmptyBytes;
        internal static readonly RandomNumberGenerator RandomNumber;
        private const string _webSocketVersion = "13";
        private const string _webSocketGuid = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11";
        private const int FRAGMENT_LENGTH = 1016;
        private const int MAX_PING_SIZE = 125;
        private const string NOT_CONNECTED_MESSAGE = "Connection state is not 'Open', please connect to the server first";
        private const string SERVER_CANNOT_USE_SET_MESSAGE = "The set operation cannot be used by servers.";
        private const string CLOSING_MESSAGE = "The connection is closing.";
        private static readonly int _maxRetryCountForConnect;
        private readonly bool _client;
        private readonly Action<MessageEventArgs> _message;
        private readonly string[] _protocols;
        private AuthenticationChallenge _authChallenge;
        private string _base64Key;
        private Action _closeContext;
        private CompressionMethod _compression;
        private WSContext _context;
        private string _extensions;
        private bool _extensionsRequested;
        private object _messageEventQueueLocker;
        private object _pingLocker;
        private object _sendLocker;
        private object _stateLocker;
        private MemoryStream _fragmentsBuffer;
        private bool _fragmentsCompressed;
        private OperationCode _fragmentsOpcode;
        private bool _inContinuation;
        private volatile bool _inMessage;
        private bool _isRedirectionEnabled;
        private Queue<MessageEventArgs> _messageEventQueue;
        private uint _nonceCount;
        private string _origin;
        private ManualResetEvent _pongReceived;
        private bool _preAuth;
        private string _protocol;
        private bool _protocolsRequested;
        private NetworkCredential _proxyCredentials;
        private Uri _proxyUri;
        private volatile WSState _readyState;
        private ManualResetEvent _receivingExited;
        private int _retryCountForConnect;
        private SSLConfigClient _sslConfig;
        private Stream _stream;
        private TcpClient _tcpClient;
        private Uri _uri;
        private TimeSpan _responseWaitingTime;

        static WSClient()
        {
            _maxRetryCountForConnect = 10;
            EmptyBytes = new byte[0];
            FragmentLength = FRAGMENT_LENGTH;
            RandomNumber = new RNGCryptoServiceProvider();
        }

        public WSClient(string url, X509Certificate clientCertificate = null, params string[] protocols)
        {
            if (url == null)
            {
                throw new ArgumentNullException(nameof(url));
            }

            if (url.Length == 0)
            {
                throw new ArgumentException("Empty string.", nameof(url));
            }

            if (!url.TryCreateWebSocketUri(out _uri, out var message))
            {
                throw new ArgumentException(message, nameof(url));
            }

            if (protocols != null && protocols.Length > 0 && !checkProtocols(protocols, out message))
            {
                throw new ArgumentException(message, nameof(protocols));
            }

            _protocols = protocols;
            _base64Key = CreateBase64Key();
            _client = true;
            _message = messagec;
            IsSSL = _uri.Scheme == "wss";
            _responseWaitingTime = TimeSpan.FromSeconds(5);

            if (IsSSL && clientCertificate != null)
            {
                SSL.Certificates.Add(clientCertificate);
            }

            Setup();
        }

        internal WSClient(HttpListenerWSContext context, string protocol)
        {
            _context = context;
            _protocol = protocol;

            _closeContext = context.Close;
            _message = messages;
            IsSSL = context.IsSecureConnection;
            _stream = context.Stream;
            _responseWaitingTime = TimeSpan.FromSeconds(1);

            Setup();
        }

        internal WSClient(TcpListenerWSContext context, string protocol)
        {
            _context = context;
            _protocol = protocol;

            _closeContext = context.Close;
            _message = messages;
            IsSSL = context.IsSecureConnection;
            _stream = context.Stream;
            _responseWaitingTime = TimeSpan.FromSeconds(1);

            Setup();
        }

        public event EventHandler OnConnect;

        public event EventHandler<CloseEventArgs> OnDisconnect;

        public event EventHandler<ErrorEventArgs> OnError;

        public event EventHandler<MessageEventArgs> OnMessageReceived;

        /// <summary>
        /// Fragment length
        /// </summary>
        public static int FragmentLength { get; set; }

        public bool CallMessageOnPing { get; set; }

        public CompressionMethod Compression
        {
            get
            {
                return _compression;
            }

            set
            {
                if (!_client)
                {
                    throw new InvalidOperationException(SERVER_CANNOT_USE_SET_MESSAGE);
                }

                if (!canSet(out string message))
                {
                    Logger.Warning(message);
                    return;
                }

                lock (_stateLocker)
                {
                    if (!canSet(out message))
                    {
                        Logger.Warning(message);
                        return;
                    }

                    _compression = value;
                }
            }
        }

        public IEnumerable<Cookie> Cookies
        {
            get
            {
                lock (CookieCollection.SyncRoot)
                {
                    foreach (Cookie cookie in CookieCollection)
                    {
                        yield return cookie;
                    }
                }
            }
        }

        public NetworkCredential Credentials { get; private set; }

        /// <summary>
        /// Gets or sets the custom headers
        /// </summary>
        public IEnumerable<KeyValuePair<string, string>> CustomHeaders { get; set; }

        public string Extensions => _extensions ?? string.Empty;
        public bool IsAlive => ping(EmptyBytes);

        public bool IsRedirectionEnabled
        {
            get
            {
                return _isRedirectionEnabled;
            }

            set
            {

                if (!_client)
                {
                    throw new InvalidOperationException(SERVER_CANNOT_USE_SET_MESSAGE);
                }

                if (!canSet(out string message))
                {
                    Logger.Warning(message);
                    return;
                }

                lock (_stateLocker)
                {
                    if (!canSet(out message))
                    {
                        Logger.Warning(message);
                        return;
                    }

                    _isRedirectionEnabled = value;
                }
            }
        }

        public bool IsSSL { get; private set; }

        public string Origin
        {
            get
            {
                return _origin;
            }

            set
            {
                string message = null;

                if (!_client)
                {
                    message = "This instance is not a client.";
                    throw new InvalidOperationException(message);
                }

                if (!value.IsNullOrEmpty())
                {
                    if (!Uri.TryCreate(value, UriKind.Absolute, out Uri uri))
                    {
                        message = "Not an absolute URI string.";
                        throw new ArgumentException(message, value);
                    }

                    if (uri.Segments.Length > 1)
                    {
                        message = "It includes the path segments.";
                        throw new ArgumentException(message, value);
                    }
                }

                if (!canSet(out message))
                {
                    Logger.Warning(message);
                    return;
                }

                lock (_stateLocker)
                {
                    if (!canSet(out message))
                    {
                        Logger.Warning(message);
                        return;
                    }

                    _origin = !value.IsNullOrEmpty() ? value.TrimEnd('/') : value;
                }
            }
        }

        public string Protocol
        {
            get
            {
                return _protocol ?? string.Empty;
            }

            internal set
            {
                _protocol = value;
            }
        }

        public WSState ReadyState => _readyState;

        public SSLConfigClient SSL
        {
            get
            {
                if (!_client)
                {
                    var message = "This instance is not a client.";
                    throw new InvalidOperationException(message);
                }

                if (!IsSSL)
                {
                    var message = "This instance does not use a secure connection.";
                    throw new InvalidOperationException(message);
                }

                _sslConfig ??= new SSLConfigClient(_uri.DnsSafeHost);

                return _sslConfig;
            }
        }

        /// <summary>
        /// Gets the TcpClient used for the WebSocket.
        /// </summary>
        public TcpClient TcpClient
        {
            get
            {
                return _tcpClient;
            }
        }

        public Uri Url => _client ? _uri : _context.RequestUri;

        internal TimeSpan ResponseWaitingTime
        {
            get
            {
                return _responseWaitingTime;
            }

            set
            {
                if (value <= TimeSpan.Zero)
                {
                    throw new ArgumentOutOfRangeException(nameof(value), "Zero or less.");
                }

                if (!canSet(out string message))
                {
                    Logger.Warning(message);
                    return;
                }

                lock (_stateLocker)
                {
                    if (!canSet(out message))
                    {
                        Logger.Warning(message);
                        return;
                    }

                    _responseWaitingTime = value;
                }
            }
        }

        internal CookieCollection CookieCollection { get; private set; }
        internal Func<WSContext, string> CustomHandshakeRequestChecker { get; set; }

        internal bool HasMessage
        {
            get
            {
                lock (_messageEventQueueLocker)
                {
                    return _messageEventQueue.Count > 0;
                }
            }
        }

        internal bool IgnoreExtensions { get; set; }

        internal bool IsConnected => _readyState == WSState.Open || _readyState == WSState.Closing;

        public void Accept()
        {
            if (!checkIfAvailable(false, true, true, false, false, false, out string message))
            {
                Logger.Error(message);
                Error("An error has occurred in accepting.", null);

                return;
            }

            if (accept())
            {
                open();
            }
        }

        public void AcceptAsync()
        {
            if (!checkIfAvailable(false, true, true, false, false, false, out string message))
            {
                Logger.Error(message);
                Error("An error has occurred in accepting.", null);
                return;
            }

            Func<bool> acceptor = accept;
            Task.Run(() =>
            {
                if (acceptor())
                {
                    open();
                }
            });
        }

        public void Close()
        {
            close((ushort)CloseStatusCode.NoStatus, string.Empty);
        }

        public void Close(ushort code)
        {
            if (!code.IsCloseStatusCode())
            {
                var message = "Less than 1000 or greater than 4999.";
                throw new ArgumentOutOfRangeException(nameof(code), message);
            }

            if (_client && code == (ushort)CloseStatusCode.ServerError)
            {
                var message = $"{(ushort)CloseStatusCode.ServerError} cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (!_client && code == (ushort)CloseStatusCode.MissingExtension)
            {
                var message = $"{(ushort)CloseStatusCode.MissingExtension} cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            close(code, string.Empty);
        }

        public void Close(CloseStatusCode code)
        {
            if (_client && code == CloseStatusCode.ServerError)
            {
                var message = "ServerError cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (!_client && code == CloseStatusCode.MissingExtension)
            {
                var message = "MandatoryExtension cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            close((ushort)code, string.Empty);
        }

        public void Close(ushort code, string reason)
        {
            if (!code.IsCloseStatusCode())
            {
                var message = "Less than 1000 or greater than 4999.";
                throw new ArgumentOutOfRangeException(nameof(code), message);
            }

            if (_client && code == (ushort)CloseStatusCode.ServerError)
            {
                var message = $"{(ushort)CloseStatusCode.ServerError} cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (!_client && code == (ushort)CloseStatusCode.MissingExtension)
            {
                var message = $"{(ushort)CloseStatusCode.MissingExtension} cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (reason.IsNullOrEmpty())
            {
                close(code, string.Empty);
                return;
            }

            if (code == (ushort)CloseStatusCode.NoStatus)
            {
                var message = $"{(ushort)CloseStatusCode.NoStatus} cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (!reason.TryGetUTF8EncodedBytes(out byte[] bytes))
            {
                var message = "It could not be UTF-8-encoded.";
                throw new ArgumentException(message, nameof(reason));
            }

            if (bytes.Length > 123)
            {
                var message = "Its size is greater than 123 bytes.";
                throw new ArgumentOutOfRangeException(nameof(reason), message);
            }

            close(code, reason);
        }

        public void Close(CloseStatusCode code, string reason)
        {
            if (_client && code == CloseStatusCode.ServerError)
            {
                var message = "ServerError cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (!_client && code == CloseStatusCode.MissingExtension)
            {
                var message = "MandatoryExtension cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (reason.IsNullOrEmpty())
            {
                close((ushort)code, string.Empty);
                return;
            }

            if (code == CloseStatusCode.NoStatus)
            {
                var message = "NoStatus cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (!reason.TryGetUTF8EncodedBytes(out byte[] bytes))
            {
                var message = "It could not be UTF-8-encoded.";
                throw new ArgumentException(message, nameof(reason));
            }

            if (bytes.Length > 123)
            {
                var message = "Its size is greater than 123 bytes.";
                throw new ArgumentOutOfRangeException(nameof(reason), message);
            }

            close((ushort)code, reason);
        }

        public void CloseAsync()
        {
            closeAsync((ushort)CloseStatusCode.NoStatus, string.Empty);
        }

        public void CloseAsync(ushort code)
        {
            if (!code.IsCloseStatusCode())
            {
                var message = "Less than 1000 or greater than 4999.";
                throw new ArgumentOutOfRangeException(nameof(code), message);
            }

            if (_client && code == (ushort)CloseStatusCode.ServerError)
            {
                var message = $"{(ushort)CloseStatusCode.ServerError} cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (!_client && code == (ushort)CloseStatusCode.MissingExtension)
            {
                var message = $"{(ushort)CloseStatusCode.MissingExtension} cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            closeAsync(code, string.Empty);
        }

        public void CloseAsync(CloseStatusCode code)
        {
            if (_client && code == CloseStatusCode.ServerError)
            {
                var message = "ServerError cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (!_client && code == CloseStatusCode.MissingExtension)
            {
                var message = "MandatoryExtension cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            closeAsync((ushort)code, string.Empty);
        }

        public void CloseAsync(ushort code, string reason)
        {
            if (!code.IsCloseStatusCode())
            {
                var message = "Less than 1000 or greater than 4999.";
                throw new ArgumentOutOfRangeException(nameof(code), message);
            }

            if (_client && code == (ushort)CloseStatusCode.ServerError)
            {
                var message = $"{(ushort)CloseStatusCode.ServerError} cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (!_client && code == (ushort)CloseStatusCode.MissingExtension)
            {
                var message = $"{(ushort)CloseStatusCode.MissingExtension} cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (reason.IsNullOrEmpty())
            {
                closeAsync(code, string.Empty);
                return;
            }

            if (code == (ushort)CloseStatusCode.NoStatus)
            {
                var message = $"{(ushort)CloseStatusCode.NoStatus} cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (!reason.TryGetUTF8EncodedBytes(out byte[] bytes))
            {
                var message = "It could not be UTF-8-encoded.";
                throw new ArgumentException(message, nameof(reason));
            }

            if (bytes.Length > 123)
            {
                var message = "Its size is greater than 123 bytes.";
                throw new ArgumentOutOfRangeException(nameof(reason), message);
            }

            closeAsync(code, reason);
        }

        public void CloseAsync(CloseStatusCode code, string reason)
        {
            if (_client && code == CloseStatusCode.ServerError)
            {
                var message = "ServerError cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (!_client && code == CloseStatusCode.MissingExtension)
            {
                var message = "MandatoryExtension cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (reason.IsNullOrEmpty())
            {
                closeAsync((ushort)code, string.Empty);
                return;
            }

            if (code == CloseStatusCode.NoStatus)
            {
                var message = "NoStatus cannot be used.";
                throw new ArgumentException(message, nameof(code));
            }

            if (!reason.TryGetUTF8EncodedBytes(out byte[] bytes))
            {
                var message = "It could not be UTF-8-encoded.";
                throw new ArgumentException(message, nameof(reason));
            }

            if (bytes.Length > 123)
            {
                var message = "Its size is greater than 123 bytes.";
                throw new ArgumentOutOfRangeException(nameof(reason), message);
            }

            closeAsync((ushort)code, reason);
        }

        /// <summary>
        /// Connect to the server
        /// </summary>
        public void Connect()
        {
            if (TryConnect(out string errorMessage))
            {
                Logger.Error(errorMessage);
                Error("An error has occurred in connecting.", null);
                return;
            }

            open();
        }

        /// <summary>
        /// COnnect to the server
        /// </summary>
        /// <returns></returns>
        public async Task ConnectAsync()
        {
            if (TryConnect(out string errorMessage))
            {
                Logger.Error(errorMessage);
                Error("An error has occurred in connecting.", null);
                return;
            }

            await Task.Run(() => open()).ConfigureAwait(false);
        }

        private bool TryConnect(out string errorMessage)
        {
            if (!checkIfAvailable(true, false, true, false, false, true, out errorMessage))
            {
                return true;
            }

            return !connect();
        }

        void IDisposable.Dispose()
        {
            close((ushort)CloseStatusCode.Away, string.Empty);
        }

        public bool Ping()
        {
            return ping(EmptyBytes);
        }

        public bool Ping(string message)
        {
            if (message.IsNullOrEmpty())
            {
                return ping(EmptyBytes);
            }

            if (!message.TryGetUTF8EncodedBytes(out byte[] bytes))
            {
                var pingMessage = "It could not be UTF-8-encoded.";
                throw new ArgumentException(pingMessage, nameof(message));
            }

            if (bytes.Length > MAX_PING_SIZE)
            {
                var pingMessage = $"Size is greater than {MAX_PING_SIZE} bytes.";
                throw new ArgumentOutOfRangeException(nameof(message), pingMessage);
            }

            return ping(bytes);
        }

        /// <summary>
        /// Sends a ping using the WebSocket connection.
        /// </summary>
        /// <returns>
        /// <c>true</c> if the send has done with no error and a pong has been
        /// received within a time; otherwise, <c>false</c>.
        /// </returns>
        public void PingAsync(Action<bool> completed)
        {
            pingAsync(EmptyBytes, completed);
        }

        public void Send(byte[] data)
        {
            if (_readyState != WSState.Open)
            {
                throw new InvalidOperationException(NOT_CONNECTED_MESSAGE);
            }

            if (data == null)
            {
                throw new ArgumentNullException(nameof(data));
            }

            send(OperationCode.Binary, new MemoryStream(data));
        }

        public void Send(FileInfo fileInfo)
        {
            if (_readyState != WSState.Open)
            {
                throw new InvalidOperationException(NOT_CONNECTED_MESSAGE);
            }

            if (fileInfo == null)
            {
                throw new ArgumentNullException(nameof(fileInfo));
            }

            if (!fileInfo.Exists)
            {
                var message = "The file does not exist.";
                throw new ArgumentException(message, nameof(fileInfo));
            }

            if (!fileInfo.TryOpenRead(out FileStream stream))
            {
                var message = "The file could not be opened.";
                throw new ArgumentException(message, nameof(fileInfo));
            }

            send(OperationCode.Binary, stream);
        }

        public void Send(string data)
        {
            if (_readyState != WSState.Open)
            {
                throw new InvalidOperationException(NOT_CONNECTED_MESSAGE);
            }

            if (data == null)
            {
                throw new ArgumentNullException(nameof(data));
            }

            if (!data.TryGetUTF8EncodedBytes(out byte[] bytes))
            {
                var message = "It could not be UTF-8-encoded.";
                throw new ArgumentException(message, nameof(data));
            }

            send(OperationCode.Text, new MemoryStream(bytes));
        }

        public void Send(Stream stream, int length)
        {
            if (_readyState != WSState.Open)
            {
                throw new InvalidOperationException(NOT_CONNECTED_MESSAGE);
            }

            if (stream == null)
            {
                throw new ArgumentNullException(nameof(stream));
            }

            if (!stream.CanRead)
            {
                var message = "It cannot be read.";
                throw new ArgumentException(message, nameof(stream));
            }

            if (length < 1)
            {
                var message = "Less than 1.";
                throw new ArgumentException(message, nameof(length));
            }

            var bytes = stream.ReadBytes(length);

            var len = bytes.Length;
            if (len == 0)
            {
                var message = "No data could be read from it.";
                throw new ArgumentException(message, nameof(stream));
            }

            if (len < length)
            {
                Logger.Warning(
                  string.Format(
                    "Only {0} byte(s) of data could be read from the stream.",
                    len
                  )
                );
            }

            send(OperationCode.Binary, new MemoryStream(bytes));
        }

        public void SendAsync(byte[] data, Action<bool> completed = null)
        {
            if (_readyState != WSState.Open)
            {
                throw new InvalidOperationException(NOT_CONNECTED_MESSAGE);
            }

            if (data == null)
            {
                throw new ArgumentNullException(nameof(data));
            }

            sendAsync(OperationCode.Binary, new MemoryStream(data), completed);
        }

        public void SendAsync(FileInfo fileInfo, Action<bool> completed)
        {
            if (_readyState != WSState.Open)
            {
                throw new InvalidOperationException(NOT_CONNECTED_MESSAGE);
            }

            if (fileInfo == null)
            {
                throw new ArgumentNullException(nameof(fileInfo));
            }

            if (!fileInfo.Exists)
            {
                var message = "The file does not exist.";
                throw new ArgumentException(message, nameof(fileInfo));
            }

            if (!fileInfo.TryOpenRead(out FileStream stream))
            {
                var message = "The file could not be opened.";
                throw new ArgumentException(message, nameof(fileInfo));
            }

            sendAsync(OperationCode.Binary, stream, completed);
        }

        public void SendAsync(string data, Action<bool> completed)
        {
            if (_readyState != WSState.Open)
            {
                throw new InvalidOperationException(NOT_CONNECTED_MESSAGE);
            }

            if (data == null)
            {
                throw new ArgumentNullException(nameof(data));
            }

            if (!data.TryGetUTF8EncodedBytes(out byte[] bytes))
            {
                var message = "It could not be UTF-8-encoded.";
                throw new ArgumentException(message, nameof(data));
            }

            sendAsync(OperationCode.Text, new MemoryStream(bytes), completed);
        }

        public void SendAsync(Stream stream, int length, Action<bool> completed)
        {
            if (_readyState != WSState.Open)
            {
                throw new InvalidOperationException(NOT_CONNECTED_MESSAGE);
            }

            if (stream == null)
            {
                throw new ArgumentNullException(nameof(stream));
            }

            if (!stream.CanRead)
            {
                var message = "It cannot be read.";
                throw new ArgumentException(message, nameof(stream));
            }

            if (length < 1)
            {
                var message = "Less than 1.";
                throw new ArgumentException(message, nameof(length));
            }

            var bytes = stream.ReadBytes(length);

            var len = bytes.Length;
            if (len == 0)
            {
                var message = "No data could be read from it.";
                throw new ArgumentException(message, nameof(stream));
            }

            if (len < length)
            {
                Logger.Warning(
                  string.Format(
                    "Only {0} byte(s) of data could be read from the stream.",
                    len
                  )
                );
            }

            sendAsync(OperationCode.Binary, new MemoryStream(bytes), completed);
        }

        public void SetCookie(Cookie cookie)
        {
            if (!checkIfAvailable(true, false, true, false, false, true, out string message))
            {
                Logger.Error(message);
                Error("An error has occurred in setting a cookie.", null);

                return;
            }

            if (cookie == null)
            {
                Logger.Error("'cookie' is null.");
                Error("An error has occurred in setting a cookie.", null);

                return;
            }

            lock (_stateLocker)
            {
                if (!checkIfAvailable(true, false, false, true, out message))
                {
                    Logger.Error(message);
                    Error("An error has occurred in setting a cookie.", null);

                    return;
                }

                lock (CookieCollection.SyncRoot)
                {
                    CookieCollection.SetOrRemove(cookie);
                }
            }
        }

        public void SetCredentials(string username, string password, bool preAuth)
        {
            if (!checkIfAvailable(true, false, true, false, false, true, out string message))
            {
                Logger.Error(message);
                Error("An error has occurred in setting the credentials.", null);

                return;
            }

            if (!checkParametersForSetCredentials(username, password, out message))
            {
                Logger.Error(message);
                Error("An error has occurred in setting the credentials.", null);

                return;
            }

            lock (_stateLocker)
            {
                if (!checkIfAvailable(true, false, false, true, out message))
                {
                    Logger.Error(message);
                    Error("An error has occurred in setting the credentials.", null);

                    return;
                }

                if (username.IsNullOrEmpty())
                {
                    Logger.Warning("The credentials are initialized.");
                    Credentials = null;
                    _preAuth = false;

                    return;
                }

                Credentials = new NetworkCredential(username, password, _uri.PathAndQuery);
                _preAuth = preAuth;
            }
        }

        public void SetProxy(string url, string username, string password)
        {
            if (!checkIfAvailable(true, false, true, false, false, true, out string message))
            {
                Logger.Error(message);
                Error("An error has occurred in setting the proxy.", null);

                return;
            }

            if (!checkParametersForSetProxy(url, username, password, out message))
            {
                Logger.Error(message);
                Error("An error has occurred in setting the proxy.", null);

                return;
            }

            lock (_stateLocker)
            {
                if (!checkIfAvailable(true, false, false, true, out message))
                {
                    Logger.Error(message);
                    Error("An error has occurred in setting the proxy.", null);

                    return;
                }

                if (url.IsNullOrEmpty())
                {
                    Logger.Warning("The url and credentials for the proxy are initialized.");
                    _proxyUri = null;
                    _proxyCredentials = null;

                    return;
                }

                _proxyUri = new Uri(url);

                if (username.IsNullOrEmpty())
                {
                    Logger.Warning("The credentials for the proxy are initialized.");
                    _proxyCredentials = null;

                    return;
                }

                _proxyCredentials =
                  new NetworkCredential(
                    username, password, string.Format("{0}:{1}", _uri.DnsSafeHost, _uri.Port)
                  );
            }
        }

        internal static string CreateBase64Key()
        {
            var src = new byte[16];
            RandomNumber.GetBytes(src);

            return Convert.ToBase64String(src);
        }

        internal static string CreateResponseKey(string base64Key)
        {
            var buff = new StringBuilder(base64Key, 64);
            buff.Append(_webSocketGuid);
            SHA1 sha1 = new SHA1CryptoServiceProvider();
            var src = sha1.ComputeHash(buff.ToString().UTF8Encode());

            return Convert.ToBase64String(src);
        }

        internal void Close(WebResponse response)
        {
            _readyState = WSState.Closing;

            sendHttpResponse(response);
            releaseServerResources();

            _readyState = WSState.Closed;
        }

        internal void Close(HttpStatusCode code)
        {
            Close(createHandshakeFailureResponse(code));
        }

        internal void Close(Payload payload, byte[] frameAsBytes)
        {
            lock (_stateLocker)
            {
                if (_readyState == WSState.Closing)
                {
                    Logger.Info("The closing is already in progress.");
                    return;
                }

                if (_readyState == WSState.Closed)
                {
                    Logger.Info("The connection has already been closed.");
                    return;
                }

                _readyState = WSState.Closing;
            }

            Logger.Trace("Begin closing the connection.");

            var sent = frameAsBytes != null && sendBytes(frameAsBytes);
            var received = sent && _receivingExited != null && _receivingExited.WaitOne(_responseWaitingTime);

            var result = sent && received;

            Logger.Debug($"Was clean?: {result}\n  sent: {sent}\n  received: {received}");

            releaseServerResources();
            releaseCommonResources();

            Logger.Trace("End closing the connection.");

            _readyState = WSState.Closed;

            var e = new CloseEventArgs(payload);
            e.WasClean = result;

            try
            {
                OnDisconnect.Emit(this, e);
            }
            catch (Exception ex)
            {
                Logger.Error(ex.ToString());
            }
        }

        internal void InternalAccept()
        {
            try
            {
                if (!acceptHandshake())
                {
                    return;
                }

                _readyState = WSState.Open;
            }
            catch (Exception ex)
            {
                Logger.Error(ex.ToString());
                fatal("An exception has occurred while accepting.", ex);

                return;
            }

            open();
        }

        internal bool Ping(byte[] frameAsBytes, TimeSpan timeout)
        {
            if (_readyState != WSState.Open)
            {
                return false;
            }

            var pongReceived = _pongReceived;
            if (pongReceived == null)
            {
                return false;
            }

            lock (_pingLocker)
            {
                try
                {
                    pongReceived.Reset();

                    lock (_stateLocker)
                    {
                        if (_readyState != WSState.Open)
                        {
                            return false;
                        }

                        if (!sendBytes(frameAsBytes))
                        {
                            return false;
                        }
                    }

                    return pongReceived.WaitOne(timeout);
                }
                catch (ObjectDisposedException)
                {
                    return false;
                }
            }
        }

        internal void Send(
          OperationCode opcode, byte[] data, Dictionary<CompressionMethod, byte[]> cache
        )
        {
            lock (_sendLocker)
            {
                lock (_stateLocker)
                {
                    if (_readyState != WSState.Open)
                    {
                        Logger.Error(CLOSING_MESSAGE);
                        return;
                    }

                    if (!cache.TryGetValue(_compression, out byte[] found))
                    {
                        found = new WSFrame(
                                  FinalFrame.Final,
                                  opcode,
                                  data.Compress(_compression),
                                  _compression != CompressionMethod.None,
                                  false
                                )
                                .ToArray();

                        cache.Add(_compression, found);
                    }

                    sendBytes(found);
                }
            }
        }

        internal void Send(
          OperationCode opcode, Stream stream, Dictionary<CompressionMethod, Stream> cache
        )
        {
            lock (_sendLocker)
            {
                if (!cache.TryGetValue(_compression, out Stream found))
                {
                    found = stream.Compress(_compression);
                    cache.Add(_compression, found);
                }
                else
                {
                    found.Position = 0;
                }

                send(opcode, found, _compression != CompressionMethod.None);
            }
        }

        private static bool checkParametersForSetCredentials(
          string username, string password, out string message
        )
        {
            message = null;

            if (username.IsNullOrEmpty())
            {
                return true;
            }

            if (username.Contains(':') || !username.IsText())
            {
                message = "'username' contains an invalid character.";
                return false;
            }

            if (password.IsNullOrEmpty())
            {
                return true;
            }

            if (!password.IsText())
            {
                message = "'password' contains an invalid character.";
                return false;
            }

            return true;
        }

        private static bool checkParametersForSetProxy(
          string url, string username, string password, out string message
        )
        {
            message = null;

            if (url.IsNullOrEmpty())
            {
                return true;
            }

            if (!Uri.TryCreate(url, UriKind.Absolute, out Uri uri)
                || uri.Scheme != "http"
                || uri.Segments.Length > 1
            )
            {
                message = "'url' is an invalid URL.";
                return false;
            }

            if (username.IsNullOrEmpty())
            {
                return true;
            }

            if (username.Contains(':') || !username.IsText())
            {
                message = "'username' contains an invalid character.";
                return false;
            }

            if (password.IsNullOrEmpty())
            {
                return true;
            }

            if (!password.IsText())
            {
                message = "'password' contains an invalid character.";
                return false;
            }

            return true;
        }

        private static bool checkProtocols(string[] protocols, out string message)
        {
            message = null;

            Func<string, bool> cond = protocol => protocol.IsNullOrEmpty()
                                                  || !protocol.IsToken();

            if (protocols.Contains(cond))
            {
                message = "It contains a value that is not a token.";
                return false;
            }

            if (protocols.ContainsTwice())
            {
                message = "It contains a value twice.";
                return false;
            }

            return true;
        }

        private bool accept()
        {
            lock (_stateLocker)
            {
                if (!checkIfAvailable(true, false, false, false, out string message))
                {
                    Logger.Error(message);
                    Error("An error has occurred in accepting.", null);

                    return false;
                }

                try
                {
                    if (!acceptHandshake())
                    {
                        return false;
                    }

                    _readyState = WSState.Open;
                }
                catch (Exception ex)
                {
                    Logger.Error(ex.ToString());
                    fatal("An exception has occurred while accepting.", ex);

                    return false;
                }

                return true;
            }
        }

        private bool acceptHandshake()
        {
            Logger.Debug($"A request from {_context.UserEndPoint}:\n{_context}");

            if (!checkHandshakeRequest(_context, out string message))
            {
                sendHttpResponse(createHandshakeFailureResponse(HttpStatusCode.BadRequest));

                Logger.Error(message);
                fatal("An error has occurred while accepting.", CloseStatusCode.ProtocolError);

                return false;
            }

            if (!customCheckHandshakeRequest(_context, out message))
            {
                sendHttpResponse(createHandshakeFailureResponse(HttpStatusCode.BadRequest));

                Logger.Error(message);
                fatal("An error has occurred while accepting.", CloseStatusCode.PolicyViolation);

                return false;
            }

            _base64Key = _context.Headers["Sec-WebSocket-Key"];

            if (_protocol != null)
            {
                processSecWebSocketProtocolHeader(_context.SecWebSocketProtocols);
            }

            if (!IgnoreExtensions)
            {
                processSecWebSocketExtensionsClientHeader(_context.Headers["Sec-WebSocket-Extensions"]);
            }

            return sendHttpResponse(createHandshakeResponse());
        }

        private bool canSet(out string message)
        {
            message = null;

            if (_readyState == WSState.Open)
            {
                message = "The connection has already been established.";
                return false;
            }

            if (_readyState == WSState.Closing)
            {
                message = CLOSING_MESSAGE;
                return false;
            }

            return true;
        }

        private bool checkHandshakeRequest(WSContext context, out string message)
        {
            message = null;

            if (context.RequestUri == null)
            {
                message = "Specifies an invalid Request-URI.";
                return false;
            }

            if (!context.IsWebSocketRequest)
            {
                message = "Not a WebSocket handshake request.";
                return false;
            }

            var headers = context.Headers;
            if (!validateSecWebSocketKeyHeader(headers["Sec-WebSocket-Key"]))
            {
                message = "Includes no Sec-WebSocket-Key header, or it has an invalid value.";
                return false;
            }

            if (!validateSecWebSocketVersionClientHeader(headers["Sec-WebSocket-Version"]))
            {
                message = "Includes no Sec-WebSocket-Version header, or it has an invalid value.";
                return false;
            }

            if (!validateSecWebSocketProtocolClientHeader(headers["Sec-WebSocket-Protocol"]))
            {
                message = "Includes an invalid Sec-WebSocket-Protocol header.";
                return false;
            }

            if (!IgnoreExtensions
                && !validateSecWebSocketExtensionsClientHeader(headers["Sec-WebSocket-Extensions"])
            )
            {
                message = "Includes an invalid Sec-WebSocket-Extensions header.";
                return false;
            }

            return true;
        }

        private bool checkHandshakeResponse(WebResponse response, out string message)
        {
            message = null;

            if (response.IsRedirect)
            {
                message = "Indicates the redirection.";
                return false;
            }

            if (response.IsUnauthorized)
            {
                message = "Requires the authentication.";
                return false;
            }

            if (!response.IsWebSocketResponse)
            {
                message = "Not a WebSocket handshake response.";
                return false;
            }

            var headers = response.Headers;
            if (!validateSecWebSocketAcceptHeader(headers["Sec-WebSocket-Accept"]))
            {
                message = "Includes no Sec-WebSocket-Accept header, or it has an invalid value.";
                return false;
            }

            if (!validateSecWebSocketProtocolServerHeader(headers["Sec-WebSocket-Protocol"]))
            {
                message = "Includes no Sec-WebSocket-Protocol header, or it has an invalid value.";
                return false;
            }

            if (!validateSecWebSocketExtensionsServerHeader(headers["Sec-WebSocket-Extensions"]))
            {
                message = "Includes an invalid Sec-WebSocket-Extensions header.";
                return false;
            }

            if (!validateSecWebSocketVersionServerHeader(headers["Sec-WebSocket-Version"]))
            {
                message = "Includes an invalid Sec-WebSocket-Version header.";
                return false;
            }

            return true;
        }

        private bool checkIfAvailable(
          bool connecting, bool open, bool closing, bool closed, out string message
        )
        {
            message = null;

            if (!connecting && _readyState == WSState.Connecting)
            {
                message = "This operation is not available in: connecting";
                return false;
            }

            if (!open && _readyState == WSState.Open)
            {
                message = "This operation is not available in: open";
                return false;
            }

            if (!closing && _readyState == WSState.Closing)
            {
                message = "This operation is not available in: closing";
                return false;
            }

            if (!closed && _readyState == WSState.Closed)
            {
                message = "This operation is not available in: closed";
                return false;
            }

            return true;
        }

        private bool checkIfAvailable(
          bool client,
          bool server,
          bool connecting,
          bool open,
          bool closing,
          bool closed,
          out string message
        )
        {
            message = null;

            if (!client && _client)
            {
                message = "This operation is not available in: client";
                return false;
            }

            if (!server && !_client)
            {
                message = "This operation is not available in: server";
                return false;
            }

            return checkIfAvailable(connecting, open, closing, closed, out message);
        }

        private bool checkReceivedFrame(WSFrame frame, out string message)
        {
            message = null;

            var masked = frame.IsMasked;
            if (_client && masked)
            {
                message = "A frame from the server is masked.";
                return false;
            }

            if (!_client && !masked)
            {
                message = "A frame from a client is not masked.";
                return false;
            }

            if (_inContinuation && frame.IsData)
            {
                message = "A data frame has been received while receiving continuation frames.";
                return false;
            }

            if (frame.IsCompressed && _compression == CompressionMethod.None)
            {
                message = "A compressed frame has been received without any agreement for it.";
                return false;
            }

            if (frame.Opcode == OperationCode.Continue && !_inContinuation)
            {
                message = "A continuation frame has been received but there is nothing to continue.";
                return false;
            }

            if (frame.Rsv2 == ReservedBits.On)
            {
                message = "The RSV2 of a frame is non-zero without any negotiation for it.";
                return false;
            }

            if (frame.Rsv3 == ReservedBits.On)
            {
                message = "The RSV3 of a frame is non-zero without any negotiation for it.";
                return false;
            }

            return true;
        }

        private void close(ushort code, string reason)
        {
            if (_readyState == WSState.Closing)
            {
                Logger.Info("The closing is already in progress.");
                return;
            }

            if (_readyState == WSState.Closed)
            {
                Logger.Info("The connection has already been closed.");
                return;
            }

            if (code == (ushort)CloseStatusCode.NoStatus)
            { // == no status
                close(Payload.Empty, true, true, false);
                return;
            }

            var send = !code.IsReserved();
            close(new Payload(code, reason), send, send, false);
        }

        private void close(Payload payload, bool send, bool receive, bool received)
        {
            lock (_stateLocker)
            {
                if (_readyState == WSState.Closing)
                {
                    Logger.Info("The closing is already in progress.");
                    return;
                }

                if (_readyState == WSState.Closed)
                {
                    Logger.Info("The connection has already been closed.");
                    return;
                }

                send = send && _readyState == WSState.Open;
                receive = send && receive;

                _readyState = WSState.Closing;
            }

            Logger.Trace("Begin closing the connection.");

            var res = closeHandshake(payload, send, receive, received);
            releaseResources();

            Logger.Trace("End closing the connection.");

            _readyState = WSState.Closed;

            var e = new CloseEventArgs(payload);
            e.WasClean = res;

            try
            {
                OnDisconnect.Emit(this, e);
            }
            catch (Exception ex)
            {
                Logger.Error(ex.ToString());
                Error("An error has occurred during the OnClose event.", ex);
            }
        }

        private void closeAsync(ushort code, string reason)
        {
            if (_readyState == WSState.Closing)
            {
                Logger.Info("The closing is already in progress.");
                return;
            }

            if (_readyState == WSState.Closed)
            {
                Logger.Info("The connection has already been closed.");
                return;
            }

            if (code == (ushort)CloseStatusCode.NoStatus)
            { // == no status
                closeAsync(Payload.Empty, true, true, false);
                return;
            }

            var send = !code.IsReserved();
            closeAsync(new Payload(code, reason), send, send, false);
        }

        private async Task closeAsync(Payload payload, bool send, bool receive, bool received)
        {
            await Task.Run(() => close(payload, send, receive, received));
        }

        private bool closeHandshake(
          Payload payload, bool send, bool receive, bool received
        )
        {
            var sent = false;
            if (send)
            {
                var frame = WSFrame.CreateCloseFrame(payload, _client);
                sent = sendBytes(frame.ToArray());

                if (_client)
                {
                    frame.Unmask();
                }
            }

            var wait = !received && sent && receive && _receivingExited != null;
            if (wait)
            {
                received = _receivingExited.WaitOne(_responseWaitingTime);
            }

            var ret = sent && received;

            Logger.Debug(
              string.Format(
                "Was clean?: {0}\n  sent: {1}\n  received: {2}", ret, sent, received
              )
            );

            return ret;
        }

        private bool connect()
        {
            lock (_stateLocker)
            {
                if (!checkIfAvailable(true, false, false, true, out string message))
                {
                    Logger.Error(message);
                    Error("An error has occurred in connecting.", null);

                    return false;
                }

                if (_retryCountForConnect > _maxRetryCountForConnect)
                {
                    _retryCountForConnect = 0;
                    Logger.Error("A series of reconnecting has failed.");

                    return false;
                }

                _readyState = WSState.Connecting;

                try
                {
                    Handshake();
                }
                catch (Exception ex)
                {
                    _retryCountForConnect++;
                    Logger.Error(ex.ToString());
                    fatal("An exception has occurred while connecting.", ex);

                    return false;
                }

                _retryCountForConnect = 1;
                _readyState = WSState.Open;

                return true;
            }
        }

        private string createExtensions()
        {
            var buff = new StringBuilder(80);

            if (_compression != CompressionMethod.None)
            {
                var str = _compression.ToExtensionString("server_no_context_takeover", "client_no_context_takeover");
                buff.AppendFormat("{0}, ", str);
            }

            var len = buff.Length;
            if (len > 2)
            {
                buff.Length = len - 2;
                return buff.ToString();
            }

            return null;
        }

        private WebResponse createHandshakeFailureResponse(HttpStatusCode code)
        {
            var result = WebResponse.CreateCloseResponse(code);
            result.Headers["Sec-WebSocket-Version"] = _webSocketVersion;
            return result;
        }

        private WebRequest createHandshakeRequest()
        {
            var ret = WebRequest.CreateWebSocketRequest(_uri);

            var headers = ret.Headers;
            if (!_origin.IsNullOrEmpty())
            {
                headers["Origin"] = _origin;
            }

            headers["Sec-WebSocket-Key"] = _base64Key;

            _protocolsRequested = _protocols != null && _protocols.Length > 0;
            if (_protocolsRequested)
            {
                headers["Sec-WebSocket-Protocol"] = _protocols.ToString(", ");
            }

            _extensionsRequested = _compression != CompressionMethod.None;
            if (_extensionsRequested)
            {
                headers["Sec-WebSocket-Extensions"] = createExtensions();
            }

            headers["Sec-WebSocket-Version"] = _webSocketVersion;

            AuthenticationResponse authRes = null;
            if (_authChallenge != null && Credentials != null)
            {
                authRes = new AuthenticationResponse(_authChallenge, Credentials, _nonceCount);
                _nonceCount = authRes.NonceCount;
            }
            else if (_preAuth)
            {
                authRes = new AuthenticationResponse(Credentials);
            }

            if (authRes != null)
            {
                headers["Authorization"] = authRes.ToString();
            }

            if (CookieCollection.Count > 0)
            {
                ret.SetCookies(CookieCollection);
            }

            if (CustomHeaders != null)
                foreach (var header in CustomHeaders)
                    if (!headers.Contains(header.Key))
                        ret.Headers[header.Key] = header.Value;

            return ret;
        }

        private WebResponse createHandshakeResponse()
        {
            var ret = WebResponse.CreateWebSocketResponse();

            var headers = ret.Headers;
            headers["Sec-WebSocket-Accept"] = CreateResponseKey(_base64Key);

            if (_protocol != null)
            {
                headers["Sec-WebSocket-Protocol"] = _protocol;
            }

            if (_extensions != null)
            {
                headers["Sec-WebSocket-Extensions"] = _extensions;
            }

            if (CookieCollection.Count > 0)
            {
                ret.SetCookies(CookieCollection);
            }

            return ret;
        }

        private bool customCheckHandshakeRequest(WSContext context, out string message)
        {
            message = null;
            return CustomHandshakeRequestChecker == null
                   || (message = CustomHandshakeRequestChecker(context)) == null;
        }

        private void EnqueueToMessageEventQueue(MessageEventArgs e)
        {
            lock (_messageEventQueueLocker)
            {
                _messageEventQueue.Enqueue(e);
            }
        }

        private void Error(string message, Exception exception)
        {
            try
            {
                OnError.Emit(this, new ErrorEventArgs(message, exception));
            }
            catch (Exception ex)
            {
                Logger.Error(ex.ToString());
            }
        }

        private void fatal(string message, Exception exception)
        {
            var code = exception is WSException
                       ? ((WSException)exception).Code
                       : CloseStatusCode.Abnormal;

            fatal(message, (ushort)code);
        }

        private void fatal(string message, ushort code)
        {
            var payload = new Payload(code, message);
            close(payload, !code.IsReserved(), false, false);
        }

        private void fatal(string message, CloseStatusCode code)
        {
            fatal(message, (ushort)code);
        }

        private void Handshake()
        {
            SetClientStream();

            var response = sendHandshakeRequest();
            if (!checkHandshakeResponse(response, out string message))
            {
                throw new WSException(CloseStatusCode.ProtocolError, message);
            }

            if (_protocolsRequested)
            {
                _protocol = response.Headers["Sec-WebSocket-Protocol"];
            }

            if (_extensionsRequested)
            {
                processSecWebSocketExtensionsServerHeader(response.Headers["Sec-WebSocket-Extensions"]);
            }

            processCookies(response.Cookies);
        }

        private void message()
        {
            MessageEventArgs e = null;
            lock (_messageEventQueueLocker)
            {
                if (_inMessage || _messageEventQueue.Count == 0 || _readyState != WSState.Open)
                {
                    return;
                }

                _inMessage = true;
                e = _messageEventQueue.Dequeue();
            }

            _message(e);
        }

        private void messagec(MessageEventArgs e)
        {
            do
            {
                try
                {
                    OnMessageReceived.Emit(this, e);
                }
                catch (Exception ex)
                {
                    Logger.Error(ex.ToString());
                    Error("An error has occurred during an OnMessage event.", ex);
                }

                lock (_messageEventQueueLocker)
                {
                    if (_messageEventQueue.Count == 0 || _readyState != WSState.Open)
                    {
                        _inMessage = false;
                        break;
                    }

                    e = _messageEventQueue.Dequeue();
                }
            }
            while (true);
        }

        private void messages(MessageEventArgs e)
        {
            try
            {
                OnMessageReceived.Emit(this, e);
            }
            catch (Exception ex)
            {
                Logger.Error(ex.ToString());
                Error("An error has occurred during an OnMessage event.", ex);
            }

            lock (_messageEventQueueLocker)
            {
                if (_messageEventQueue.Count == 0 || _readyState != WSState.Open)
                {
                    _inMessage = false;
                    return;
                }

                e = _messageEventQueue.Dequeue();
            }

            ThreadPool.QueueUserWorkItem(state => messages(e));
        }

        private void open()
        {
            _inMessage = true;
            startReceiving();
            try
            {
                OnConnect.Emit(this, EventArgs.Empty);
            }
            catch (Exception ex)
            {
                Logger.Error(ex.ToString());
                Error("An error has occurred during the OnOpen event.", ex);
            }

            MessageEventArgs e = null;
            lock (_messageEventQueueLocker)
            {
                if (_messageEventQueue.Count == 0 || _readyState != WSState.Open)
                {
                    _inMessage = false;
                    return;
                }

                e = _messageEventQueue.Dequeue();
            }

            Task.Run(() => _message(e));
        }

        private bool ping(byte[] data)
        {
            if (_readyState != WSState.Open)
            {
                return false;
            }

            var pongReceived = _pongReceived;
            if (pongReceived == null)
            {
                return false;
            }

            lock (_pingLocker)
            {
                try
                {
                    pongReceived.Reset();
                    if (!send(FinalFrame.Final, OperationCode.Ping, data, false))
                    {
                        return false;
                    }

                    return pongReceived.WaitOne(_responseWaitingTime);
                }
                catch (ObjectDisposedException)
                {
                    return false;
                }
            }
        }

        private void pingAsync(byte[] data, Action<bool> completed)
        {
            if (_readyState != WSState.Open)
            {
                throw new InvalidOperationException(NOT_CONNECTED_MESSAGE);
            }

            sendAsync(OperationCode.Ping, new MemoryStream(data), completed);
        }

        private bool processCloseFrame(WSFrame frame)
        {
            var payload = frame.Payload;
            close(payload, !payload.HasReservedCode, false, true);

            return ping(EmptyBytes);
        }

        private void processCookies(CookieCollection cookies)
        {
            if (cookies.Count == 0)
            {
                return;
            }

            CookieCollection.SetOrRemove(cookies);
        }

        private bool processDataFrame(WSFrame frame)
        {
            EnqueueToMessageEventQueue(
              frame.IsCompressed
              ? new MessageEventArgs(
                  frame.Opcode, frame.Payload.ApplicationData.Decompress(_compression))
              : new MessageEventArgs(frame));

            return true;
        }

        private bool processFragmentFrame(WSFrame frame)
        {
            if (!_inContinuation)
            {
                // Must process first fragment.
                if (frame.IsContinuation)
                {
                    return true;
                }

                _fragmentsOpcode = frame.Opcode;
                _fragmentsCompressed = frame.IsCompressed;
                _fragmentsBuffer = new MemoryStream();
                _inContinuation = true;
            }

            _fragmentsBuffer.WriteBytes(frame.Payload.ApplicationData, 1024);
            if (frame.IsFinal)
            {
                using (_fragmentsBuffer)
                {
                    var data = _fragmentsCompressed
                               ? _fragmentsBuffer.DecompressToArray(_compression)
                               : _fragmentsBuffer.ToArray();

                    EnqueueToMessageEventQueue(new MessageEventArgs(_fragmentsOpcode, data));
                }

                _fragmentsBuffer = null;
                _inContinuation = false;
            }

            return true;
        }

        private bool processPingFrame(WSFrame frame)
        {
            Logger.Trace("A ping was received.");

            var pong = WSFrame.CreatePongFrame(frame.Payload, _client);

            lock (_stateLocker)
            {
                if (_readyState != WSState.Open)
                {
                    Logger.Error(CLOSING_MESSAGE);
                    return true;
                }

                if (!sendBytes(pong.ToArray()))
                {
                    return false;
                }
            }

            Logger.Trace("A pong to this ping has been sent.");

            if (CallMessageOnPing)
            {
                if (_client)
                {
                    pong.Unmask();
                }

                EnqueueToMessageEventQueue(new MessageEventArgs(frame));
            }

            return true;
        }

        private bool processPongFrame(WSFrame frame)
        {
            Logger.Trace("A pong was received.");

            try
            {
                _pongReceived.Set();
            }
            catch (NullReferenceException ex)
            {
                Logger.Error(ex.Message);
                Logger.Debug(ex.ToString());

                return false;
            }
            catch (ObjectDisposedException ex)
            {
                Logger.Error(ex.Message);
                Logger.Debug(ex.ToString());

                return false;
            }

            Logger.Trace("It has been signaled.");

            return true;
        }

        private bool processReceivedFrame(WSFrame frame)
        {
            if (!checkReceivedFrame(frame, out string message))
            {
                throw new WSException(CloseStatusCode.ProtocolError, message);
            }

            frame.Unmask();
            return frame.IsFragment
                   ? processFragmentFrame(frame)
                   : frame.IsData
                     ? processDataFrame(frame)
                     : frame.IsPing
                       ? processPingFrame(frame)
                       : frame.IsPong
                         ? processPongFrame(frame)
                         : frame.IsClose
                           ? processCloseFrame(frame)
                           : processUnsupportedFrame(frame);
        }

        private void processSecWebSocketExtensionsClientHeader(string value)
        {
            if (value == null)
            {
                return;
            }

            var buff = new StringBuilder(80);

            var comp = false;
            foreach (var extensions in value.SplitHeaderValue(','))
            {
                var extension = extensions.Trim();
                if (!comp && extension.IsCompressionExtension(CompressionMethod.Deflate))
                {
                    _compression = CompressionMethod.Deflate;
                    buff.AppendFormat($"{_compression.ToExtensionString("client_no_context_takeover", "server_no_context_takeover")}, ");
                    comp = true;
                }
            }

            var len = buff.Length;
            if (len > 2)
            {
                buff.Length = len - 2;
                _extensions = buff.ToString();
            }
        }

        private void processSecWebSocketExtensionsServerHeader(string value)
        {
            if (value == null)
            {
                _compression = CompressionMethod.None;
                return;
            }

            _extensions = value;
        }

        private void processSecWebSocketProtocolHeader(IEnumerable<string> values)
        {
            if (values.Contains(p => p == _protocol))
            {
                return;
            }

            _protocol = null;
        }

        private bool processUnsupportedFrame(WSFrame frame)
        {
            Logger.Error("An unsupported frame:" + frame.PrintToString(false));
            fatal("There is no way to handle it.", CloseStatusCode.PolicyViolation);
            return false;
        }

        private void releaseClientResources()
        {
            if (_stream != null)
            {
                _stream.Dispose();
                _stream = null;
            }

            if (_tcpClient != null)
            {
                _tcpClient.Close();
                _tcpClient = null;
            }
        }

        private void releaseCommonResources()
        {
            if (_fragmentsBuffer != null)
            {
                _fragmentsBuffer.Dispose();
                _fragmentsBuffer = null;
                _inContinuation = false;
            }

            if (_pongReceived != null)
            {
                _pongReceived.Close();
                _pongReceived = null;
            }

            if (_receivingExited != null)
            {
                _receivingExited.Close();
                _receivingExited = null;
            }
        }

        private void releaseResources()
        {
            if (_client)
            {
                releaseClientResources();
            }
            else
            {
                releaseServerResources();
            }

            releaseCommonResources();
        }

        private void releaseServerResources()
        {
            if (_closeContext == null)
            {
                return;
            }

            _closeContext();
            _closeContext = null;
            _stream = null;
            _context = null;
        }

        private bool send(OperationCode opcode, Stream stream)
        {
            lock (_sendLocker)
            {
                var src = stream;
                var compressed = false;
                var sent = false;
                try
                {
                    if (_compression != CompressionMethod.None)
                    {
                        stream = stream.Compress(_compression);
                        compressed = true;
                    }

                    sent = send(opcode, stream, compressed);
                    if (!sent)
                    {
                        Error("A send has been interrupted.", null);
                    }
                }
                catch (Exception ex)
                {
                    Logger.Error(ex.ToString());
                    Error("An error has occurred during a send.", ex);
                }
                finally
                {
                    if (compressed)
                    {
                        stream.Dispose();
                    }

                    src.Dispose();
                }

                return sent;
            }
        }

        private bool send(OperationCode opcode, Stream stream, bool compressed)
        {
            var len = stream.Length;
            if (len == 0)
            {
                return send(FinalFrame.Final, opcode, EmptyBytes, false);
            }

            var quo = len / FragmentLength;
            var rem = (int)(len % FragmentLength);

            byte[] buff = null;
            if (quo == 0)
            {
                buff = new byte[rem];
                return stream.Read(buff, 0, rem) == rem
                       && send(FinalFrame.Final, opcode, buff, compressed);
            }

            if (quo == 1 && rem == 0)
            {
                buff = new byte[FragmentLength];
                return stream.Read(buff, 0, FragmentLength) == FragmentLength
                       && send(FinalFrame.Final, opcode, buff, compressed);
            }

            /* Send fragments */

            // Begin
            buff = new byte[FragmentLength];
            var sent = stream.Read(buff, 0, FragmentLength) == FragmentLength
                       && send(FinalFrame.More, opcode, buff, compressed);

            if (!sent)
            {
                return false;
            }

            var n = rem == 0 ? quo - 2 : quo - 1;
            for (long i = 0; i < n; i++)
            {
                sent = stream.Read(buff, 0, FragmentLength) == FragmentLength
                       && send(FinalFrame.More, OperationCode.Continue, buff, false);

                if (!sent)
                {
                    return false;
                }
            }

            // End
            if (rem == 0)
            {
                rem = FragmentLength;
            }
            else
            {
                buff = new byte[rem];
            }

            return stream.Read(buff, 0, rem) == rem
                   && send(FinalFrame.Final, OperationCode.Continue, buff, false);
        }

        private bool send(FinalFrame fin, OperationCode opcode, byte[] data, bool compressed)
        {
            lock (_stateLocker)
            {
                if (_readyState != WSState.Open)
                {
                    Logger.Error(CLOSING_MESSAGE);
                    return false;
                }

                var frame = new WSFrame(fin, opcode, data, compressed, _client);
                return sendBytes(frame.ToArray());
            }
        }

        private void sendAsync(OperationCode opcode, Stream stream, Action<bool> completed)
        {
            Func<OperationCode, Stream, bool> sender = send;
            Task.Run(() =>
            {
                try
                {
                    var sent = sender(opcode, stream);
                    completed?.Invoke(sent);
                }
                catch (Exception ex)
                {
                    Logger.Error(ex.ToString());
                    Error("An error has occurred during the callback for an async send.", ex);
                }
            });
        }

        private bool sendBytes(byte[] bytes)
        {
            try
            {
                _stream.Write(bytes, 0, bytes.Length);
            }
            catch (Exception ex)
            {
                Logger.Error(ex.Message);
                Logger.Debug(ex.ToString());

                return false;
            }

            return true;
        }

        private WebResponse sendHandshakeRequest()
        {
            var req = createHandshakeRequest();
            var result = sendHttpRequest(req, 90000);
            if (result.IsUnauthorized)
            {
                var challenge = result.Headers["WWW-Authenticate"];
                Logger.Warning($"Received an authentication requirement for '{challenge}'.");
                if (challenge.IsNullOrEmpty())
                {
                    Logger.Error("No authentication challenge is specified.");
                    return result;
                }

                _authChallenge = AuthenticationChallenge.Parse(challenge);
                if (_authChallenge == null)
                {
                    Logger.Error("An invalid authentication challenge is specified.");
                    return result;
                }

                if (Credentials != null &&
                    (!_preAuth || _authChallenge.Scheme == AuthenticationSchemes.Digest))
                {
                    if (result.HasConnectionClose)
                    {
                        releaseClientResources();
                        SetClientStream();
                    }

                    var authRes = new AuthenticationResponse(_authChallenge, Credentials, _nonceCount);
                    _nonceCount = authRes.NonceCount;
                    req.Headers["Authorization"] = authRes.ToString();
                    result = sendHttpRequest(req, 15000);
                }
            }

            if (result.IsRedirect)
            {
                var url = result.Headers["Location"];
                Logger.Warning($"Received a redirection to '{url}'.");
                if (_isRedirectionEnabled)
                {
                    if (url.IsNullOrEmpty())
                    {
                        Logger.Error("No url to redirect is located.");
                        return result;
                    }

                    if (!url.TryCreateWebSocketUri(out Uri uri, out string message))
                    {
                        Logger.Error("An invalid url to redirect is located: " + message);
                        return result;
                    }

                    releaseClientResources();

                    _uri = uri;
                    IsSSL = uri.Scheme == "wss";

                    SetClientStream();
                    return sendHandshakeRequest();
                }
            }

            return result;
        }

        private WebResponse sendHttpRequest(WebRequest request, int millisecondsTimeout)
        {
            Logger.Debug("Request to server:\n" + request.ToString());
            var res = request.GetResponse(_stream, millisecondsTimeout);
            Logger.Debug("Response to request:\n" + res.ToString());

            return res;
        }

        private bool sendHttpResponse(WebResponse response)
        {
            Logger.Debug("Response to request:\n" + response.ToString());
            return sendBytes(response.ToByteArray());
        }

        private void sendProxyConnectRequest()
        {
            var req = WebRequest.CreateConnectRequest(_uri);
            var result = sendHttpRequest(req, 90000);
            if (result.IsProxyAuthenticationRequired)
            {
                var challenge = result.Headers["Proxy-Authenticate"];
                Logger.Warning($"Received a proxy authentication requirement for '{challenge}'.");

                if (challenge.IsNullOrEmpty())
                {
                    throw new WSException("No proxy authentication challenge is specified.");
                }

                var authChallenge = AuthenticationChallenge.Parse(challenge) ?? throw new WSException("An invalid proxy authentication challenge is specified.");
                if (_proxyCredentials != null)
                {
                    if (result.HasConnectionClose)
                    {
                        releaseClientResources();
                        _tcpClient = new TcpClient(_proxyUri.DnsSafeHost, _proxyUri.Port);
                        _stream = _tcpClient.GetStream();
                    }

                    var authRes = new AuthenticationResponse(authChallenge, _proxyCredentials, 0);
                    req.Headers["Proxy-Authorization"] = authRes.ToString();
                    result = sendHttpRequest(req, 15000);
                }

                if (result.IsProxyAuthenticationRequired)
                {
                    throw new WSException("A proxy authentication is required.");
                }
            }

            if (result.StatusCode[0] != '2')
            {
                throw new WSException("The proxy has failed a connection to the requested host and port.");
            }
        }

        private void SetClientStream()
        {
            if (_proxyUri != null)
            {
                _tcpClient = new TcpClient(_proxyUri.DnsSafeHost, _proxyUri.Port);
                _stream = _tcpClient.GetStream();
                sendProxyConnectRequest();
            }
            else
            {
                _tcpClient = new TcpClient(_uri.DnsSafeHost, _uri.Port);
                _stream = _tcpClient.GetStream();
            }

            if (IsSSL)
            {
                var conf = SSL;
                var host = conf.TargetHost;
                if (host != _uri.DnsSafeHost)
                {
                    throw new WSException(
                      CloseStatusCode.TlsHandshakeFailure, "An invalid host name is specified.");
                }

                try
                {
                    var sslStream = new SslStream(
                      _stream,
                      false,
                      conf.ServerCertificateValidationCallback,
                      conf.ClientCertificateSelectionCallback);

                    sslStream.ReadTimeout = (int)ResponseWaitingTime.TotalMilliseconds;
                    sslStream.WriteTimeout = (int)ResponseWaitingTime.TotalMilliseconds;

                    sslStream.AuthenticateAsClient(
                      host,
                      conf.Certificates,
                      conf.SslProtocols,
                      conf.CheckForCertificateRevocation);

                    _stream = sslStream;
                }
                catch (Exception ex)
                {
                    throw new WSException(CloseStatusCode.TlsHandshakeFailure, ex);
                }
            }
        }

        private void Setup()
        {
            _compression = CompressionMethod.None;
            CookieCollection = new CookieCollection();
            _pingLocker = new object();
            _sendLocker = new object();
            _stateLocker = new object();
            _messageEventQueue = new Queue<MessageEventArgs>();
            _messageEventQueueLocker = ((ICollection)_messageEventQueue).SyncRoot;
            _readyState = WSState.Connecting;
        }

        private void startReceiving()
        {
            if (_messageEventQueue.Count > 0)
            {
                _messageEventQueue.Clear();
            }

            _pongReceived = new ManualResetEvent(false);
            _receivingExited = new ManualResetEvent(false);

            Action receive = null;
            receive =
              () =>
                WSFrame.ReadFrameAsync(
                  _stream,
                  false,
                  frame =>
                  {
                      if (!processReceivedFrame(frame) || _readyState == WSState.Closed)
                      {
                          var exited = _receivingExited;
                          exited?.Set();

                          return;
                      }

                      // Receive next because the Ping or Close needs a response to it.
                      receive();

                      if (_inMessage || !HasMessage || _readyState != WSState.Open)
                      {
                          return;
                      }

                      message();
                  },
                  ex =>
                  {
                      Logger.Error(ex.ToString());
                      fatal("An exception has occurred while receiving.", ex);
                  }
                );

            receive();
        }

        private bool validateSecWebSocketAcceptHeader(string value)
        {
            return value != null && value == CreateResponseKey(_base64Key);
        }

        private bool validateSecWebSocketExtensionsClientHeader(string value)
        {
            return value == null || value.Length > 0;
        }

        private bool validateSecWebSocketExtensionsServerHeader(string value)
        {
            if (value == null)
            {
                return true;
            }

            if (value.Length == 0)
            {
                return false;
            }

            if (!_extensionsRequested)
            {
                return false;
            }

            var comp = _compression != CompressionMethod.None;
            foreach (var currentExtension in value.SplitHeaderValue(','))
            {
                var extension = currentExtension.Trim();
                if (comp && extension.IsCompressionExtension(_compression))
                {
                    if (!extension.Contains("server_no_context_takeover"))
                    {
                        Logger.Error("The server hasn't sent back 'server_no_context_takeover'.");
                        return false;
                    }

                    if (!extension.Contains("client_no_context_takeover"))
                    {
                        Logger.Warning("The server hasn't sent back 'client_no_context_takeover'.");
                    }

                    var method = _compression.ToExtensionString();
                    var invalid =
                      extension.SplitHeaderValue(';').Contains(
                        x =>
                        {
                            x = x.Trim();
                            return x != method
                         && x != "server_no_context_takeover"
                         && x != "client_no_context_takeover";
                        }
                      );

                    if (invalid)
                    {
                        return false;
                    }
                }
                else
                {
                    return false;
                }
            }

            return true;
        }

        private bool validateSecWebSocketKeyHeader(string value)
        {
            return value != null && value.Length > 0;
        }

        private bool validateSecWebSocketProtocolClientHeader(string value)
        {
            return value == null || value.Length > 0;
        }

        private bool validateSecWebSocketProtocolServerHeader(string value)
        {
            if (value == null)
            {
                return !_protocolsRequested;
            }

            if (value.Length == 0)
            {
                return false;
            }

            return _protocolsRequested && _protocols.Contains(p => p == value);
        }

        private bool validateSecWebSocketVersionClientHeader(string value)
        {
            return value != null && value == _webSocketVersion;
        }

        private bool validateSecWebSocketVersionServerHeader(string value)
        {
            return value == null || value == _webSocketVersion;
        }
    }
}