// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections.Specialized;
using System.IO;
using System.Text;

namespace EonaCat.Network
{
    internal class WebRequest : WebBase
    {
        private bool _websocketRequest;
        private bool _websocketRequestSet;

        internal WebRequest(string method, string uri)
          : this(method, uri, HttpVersion.Version11, new NameValueCollection())
        {
            Headers["User-Agent"] = $"EonaCat.Network/{Constants.Version}";
        }

        private WebRequest(string method, string uri, Version version, NameValueCollection headers)
                  : base(version, headers)
        {
            HttpMethod = method;
            RequestUri = uri;
        }

        public AuthenticationResponse AuthenticationResponse
        {
            get
            {
                var res = Headers["Authorization"];
                return res != null && res.Length > 0
                       ? AuthenticationResponse.Parse(res)
                       : null;
            }
        }

        public CookieCollection Cookies => Headers.GetCookies(false);

        public string HttpMethod { get; }

        public bool IsWebSocketRequest
        {
            get
            {
                if (!_websocketRequestSet)
                {
                    var headers = Headers;
                    _websocketRequest = HttpMethod == "GET" &&
                                        ProtocolVersion > HttpVersion.Version10 &&
                                        headers.Contains("Upgrade", "websocket") &&
                                        headers.Contains("Connection", "Upgrade");

                    _websocketRequestSet = true;
                }

                return _websocketRequest;
            }
        }

        public string RequestUri { get; }

        public void SetCookies(CookieCollection cookies)
        {
            if (cookies == null || cookies.Count == 0)
            {
                return;
            }

            var buff = new StringBuilder(64);
            foreach (var cookie in cookies.Sorted)
            {
                if (!cookie.Expired)
                {
                    buff.AppendFormat("{0}; ", cookie.ToString());
                }
            }

            var len = buff.Length;
            if (len > 2)
            {
                buff.Length = len - 2;
                Headers["Cookie"] = buff.ToString();
            }
        }

        public override string ToString()
        {
            var output = new StringBuilder(64);
            output.AppendFormat("{0} {1} HTTP/{2}{3}", HttpMethod, RequestUri, ProtocolVersion, CrLf);

            var headers = Headers;
            foreach (var key in headers.AllKeys)
            {
                output.AppendFormat("{0}: {1}{2}", key, headers[key], CrLf);
            }

            output.Append(CrLf);

            var entity = EntityBody;
            if (entity.Length > 0)
            {
                output.Append(entity);
            }

            return output.ToString();
        }

        internal static WebRequest CreateConnectRequest(Uri uri)
        {
            var host = uri.DnsSafeHost;
            var port = uri.Port;
            var authority = string.Format("{0}:{1}", host, port);
            var req = new WebRequest("CONNECT", authority);
            req.Headers["Host"] = port == 80 ? host : authority;

            return req;
        }

        internal static WebRequest CreateWebSocketRequest(Uri uri)
        {
            var req = new WebRequest("GET", uri.PathAndQuery);
            var headers = req.Headers;

            // Only includes a port number in the Host header value if it's non-default.
            // See: https://tools.ietf.org/html/rfc6455#page-17
            var port = uri.Port;
            var schm = uri.Scheme;
            headers["Host"] = (port == 80 && schm == "ws") || (port == 443 && schm == "wss")
                              ? uri.DnsSafeHost
                              : uri.Authority;

            headers["Upgrade"] = "websocket";
            headers["Connection"] = "Upgrade";

            return req;
        }

        internal static WebRequest Parse(string[] headerParts)
        {
            var requestLine = headerParts[0].Split(new[] { ' ' }, 3);
            if (requestLine.Length != 3)
            {
                throw new ArgumentException("Invalid request line: " + headerParts[0]);
            }

            var headers = new WebHeaderCollection();
            for (int i = 1; i < headerParts.Length; i++)
            {
                headers.InternalSet(headerParts[i], false);
            }

            return new WebRequest(
              requestLine[0], requestLine[1], new Version(requestLine[2].Substring(5)), headers);
        }

        internal static WebRequest Read(Stream stream, int millisecondsTimeout)
        {
            return Read(stream, Parse, millisecondsTimeout);
        }

        internal WebResponse GetResponse(Stream stream, int millisecondsTimeout)
        {
            var buff = ToByteArray();
            stream.Write(buff, 0, buff.Length);

            return Read(stream, WebResponse.Parse, millisecondsTimeout);
        }
    }
}