﻿using System;
using System.IO;
using System.Threading;
using System.Threading.Tasks;

namespace EonaCat.Network
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/License for full license details.

    /// <summary>
    /// Content route handler.  Handles GET and HEAD requests to content routes for files and directories.
    /// </summary>
    public class ContentRouteHandler
    {
        /// <summary>
        /// The FileMode value to use when accessing files within a content route via a FileStream.  Default is FileMode.Open.
        /// </summary>
        public FileMode ContentFileMode = FileMode.Open;

        /// <summary>
        /// The FileAccess value to use when accessing files within a content route via a FileStream.  Default is FileAccess.Read.
        /// </summary>
        public FileAccess ContentFileAccess = FileAccess.Read;

        /// <summary>
        /// The FileShare value to use when accessing files within a content route via a FileStream.  Default is FileShare.Read.
        /// </summary>
        public FileShare ContentFileShare = FileShare.Read;

        private readonly ContentRouteManager _Routes;

        internal ContentRouteHandler(ContentRouteManager routes)
        {
            if (routes == null)
            {
                throw new ArgumentNullException(nameof(routes));
            }

            _Routes = routes;
        }

        internal async Task Process(HttpContext context, CancellationToken token)
        {
            if (context == null)
            {
                throw new ArgumentNullException(nameof(context));
            }

            if (context.Request == null)
            {
                throw new ArgumentNullException(nameof(context.Request));
            }

            if (context.Response == null)
            {
                throw new ArgumentNullException(nameof(context.Response));
            }

            if (context.Request.Method != HttpMethod.GET
                && context.Request.Method != HttpMethod.HEAD)
            {
                Set500Response(context);
                await context.Response.Send(token).ConfigureAwait(false);
                return;
            }

            string filePath = context.Request.Url.RawWithoutQuery;
            if (!string.IsNullOrEmpty(filePath))
            {
                while (filePath.StartsWith("/"))
                {
                    filePath = filePath.Substring(1);
                }
            }

            string baseDirectory = _Routes.BaseDirectory;
            baseDirectory = baseDirectory.Replace("\\", "/");
            if (!baseDirectory.EndsWith("/"))
            {
                baseDirectory += "/";
            }

            filePath = baseDirectory + filePath;
            filePath = filePath.Replace("+", " ").Replace("%20", " ");

            string contentType = GetContentType(filePath);

            if (!File.Exists(filePath))
            {
                Set404Response(context);
                await context.Response.Send(token).ConfigureAwait(false);
                return;
            }

            FileInfo fi = new FileInfo(filePath);
            long contentLength = fi.Length;

            if (context.Request.Method == HttpMethod.GET)
            {
                FileStream fs = new FileStream(filePath, ContentFileMode, ContentFileAccess, ContentFileShare);
                context.Response.StatusCode = 200;
                context.Response.ContentLength = contentLength;
                context.Response.ContentType = GetContentType(filePath);
                await context.Response.Send(contentLength, fs, token).ConfigureAwait(false);
                return;
            }
            else if (context.Request.Method == HttpMethod.HEAD)
            {
                context.Response.StatusCode = 200;
                context.Response.ContentLength = contentLength;
                context.Response.ContentType = GetContentType(filePath);
                await context.Response.Send(contentLength, token).ConfigureAwait(false);
                return;
            }
            else
            {
                Set500Response(context);
                await context.Response.Send(token).ConfigureAwait(false);
                return;
            }
        }

        private string GetContentType(string path)
        {
            if (string.IsNullOrEmpty(path))
            {
                return "application/octet-stream";
            }

            int idx = path.LastIndexOf(".");
            if (idx >= 0)
            {
                return MimeTypes.GetFromExtension(path.Substring(idx));
            }

            return "application/octet-stream";
        }

        private void Set404Response(HttpContext context)
        {
            context.Response.StatusCode = 404;
            context.Response.ContentLength = 0;
        }

        private void Set500Response(HttpContext context)
        {
            context.Response.StatusCode = 500;
            context.Response.ContentLength = 0;
        }
    }
}