﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Threading;
using System.Threading.Tasks;

namespace EonaCat.Network
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/License for full license details.

    /// <summary>
    /// EonaCat Webserver.
    /// </summary>
    public class WebServer : IDisposable
    {
        /// <summary>
        /// Indicates whether or not the server is listening.
        /// </summary>
        public bool IsListening => (_HttpListener != null) && _HttpListener.IsListening;

        /// <summary>
        /// Number of requests being serviced currently.
        /// </summary>
        public int RequestCount => _RequestCount;

        /// <summary>
        /// EonaCat webserver settings.
        /// </summary>
        public EonaCatWebserverSettings Settings
        {
            get
            {
                return _Settings;
            }
            set
            {
                if (value == null)
                {
                    _Settings = new EonaCatWebserverSettings();
                }
                else
                {
                    _Settings = value;
                }
            }
        }

        /// <summary>
        /// EonaCat webserver routes.
        /// </summary>
        public EonaCatWebserverRoutes Routes
        {
            get
            {
                return _Routes;
            }
            set
            {
                if (value == null)
                {
                    _Routes = new EonaCatWebserverRoutes();
                }
                else
                {
                    _Routes = value;
                }
            }
        }

        /// <summary>
        /// EonaCat webserver statistics.
        /// </summary>
        public EonaCatWebserverStatistics Statistics { get; private set; } = new EonaCatWebserverStatistics();

        /// <summary>
        /// Set specific actions/callbacks to use when events are raised.
        /// </summary>
        public EonaCatWebserverEvents Events { get; private set; } = new EonaCatWebserverEvents();

        /// <summary>
        /// Default pages served by the EonaCat webserver.
        /// </summary>
        public EonaCatWebserverPages Pages { get; private set; } = new EonaCatWebserverPages();

        private readonly string _Header = "[EonaCat] ";
        private readonly Assembly _Assembly = Assembly.GetCallingAssembly();
        private EonaCatWebserverSettings _Settings = new EonaCatWebserverSettings();
        private EonaCatWebserverRoutes _Routes = new EonaCatWebserverRoutes();
        private System.Net.HttpListener _HttpListener = new System.Net.HttpListener();
        private int _RequestCount = 0;

        private CancellationTokenSource _TokenSource = new CancellationTokenSource();
        private CancellationToken _Token;
        private Task _AcceptConnections = null;

        /// <summary>
        /// Creates a new instance of the EonaCat webserver.
        /// If you do not provide a settings object, default settings will be used, which will cause EonaCat Webserver to listen on http://127.0.0.1:8000, and send events to the console.
        /// </summary>
        /// <param name="settings">EonaCat webserver settings.</param>
        /// <param name="defaultRoute">Method used when a request is received and no matching routes are found.  Commonly used as the 404 handler when routes are used.</param>
        public WebServer(EonaCatWebserverSettings settings = null, Func<HttpContext, Task> defaultRoute = null)
        {
            if (settings == null)
            {
                settings = new EonaCatWebserverSettings();
                settings.Prefixes.Add("http://127.0.0.1:8000/");
                Events.Logger = Console.WriteLine;
            }

            _Settings = settings;
            _Routes.Default = defaultRoute;
        }

        /// <summary>
        /// Creates a new instance of the EonaCat webserver.
        /// </summary>
        /// <param name="hostname">Hostname or IP address on which to listen.</param>
        /// <param name="port">TCP port on which to listen.</param>
        /// <param name="ssl">Specify whether or not SSL should be used (HTTPS).</param>
        /// <param name="defaultRoute">Method used when a request is received and no matching routes are found.  Commonly used as the 404 handler when routes are used.</param>
        public WebServer(string hostname, int port, bool ssl = false, Func<HttpContext, Task> defaultRoute = null)
        {
            if (string.IsNullOrEmpty(hostname))
            {
                hostname = "localhost";
            }

            if (port < 1)
            {
                throw new ArgumentOutOfRangeException(nameof(port));
            }

            _Settings = new EonaCatWebserverSettings(hostname, port, ssl);
            _Routes.Default = defaultRoute;
        }

        /// <summary>
        /// Creates a new instance of the EonaCat webserver.
        /// </summary>
        /// <param name="hostnames">Hostnames or IP addresses on which to listen.  Note: multiple listener endpoints are not supported on all platforms.</param>
        /// <param name="port">TCP port on which to listen.</param>
        /// <param name="ssl">Specify whether or not SSL should be used (HTTPS).</param>
        /// <param name="defaultRoute">Method used when a request is received and no matching routes are found.  Commonly used as the 404 handler when routes are used.</param>
        public WebServer(List<string> hostnames, int port, bool ssl = false, Func<HttpContext, Task> defaultRoute = null)
        {
            if (hostnames == null || hostnames.Count < 1)
            {
                hostnames = new List<string> { "localhost" };
            }

            if (port < 1)
            {
                throw new ArgumentOutOfRangeException(nameof(port));
            }

            _Settings = new EonaCatWebserverSettings(hostnames, port, ssl);
            _Routes.Default = defaultRoute;
        }

        /// <summary>
        /// Tear down the server and dispose of background workers.
        /// Do not use this object after disposal.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Start accepting new connections.
        /// </summary>
        /// <param name="token">Cancellation token useful for canceling the server.</param>
        public void Start(CancellationToken token = default)
        {
            if (_HttpListener != null && _HttpListener.IsListening)
            {
                throw new InvalidOperationException("EonaCat Webserver is already listening.");
            }

            _HttpListener = new System.Net.HttpListener();

            LoadRoutes();
            Statistics = new EonaCatWebserverStatistics();

            _TokenSource = CancellationTokenSource.CreateLinkedTokenSource(token);
            _Token = token;

            foreach (string prefix in _Settings.Prefixes)
            {
                _HttpListener.Prefixes.Add(prefix);
            }

            _HttpListener.Start();
            _AcceptConnections = Task.Run(() => AcceptConnections(_Token), _Token);
        }

        /// <summary>
        /// Start accepting new connections.
        /// </summary>
        /// <param name="token">Cancellation token useful for canceling the server.</param>
        /// <returns>Task.</returns>
        public Task StartAsync(CancellationToken token = default)
        {
            if (_HttpListener != null && _HttpListener.IsListening)
            {
                throw new InvalidOperationException("EonaCat Webserver is already listening.");
            }

            _HttpListener = new System.Net.HttpListener();

            LoadRoutes();
            Statistics = new EonaCatWebserverStatistics();

            _TokenSource = CancellationTokenSource.CreateLinkedTokenSource(token);
            _Token = token;

            foreach (string prefix in _Settings.Prefixes)
            {
                _HttpListener.Prefixes.Add(prefix);
            }

            _HttpListener.Start();
            _AcceptConnections = Task.Run(() => AcceptConnections(_Token), _Token);
            return _AcceptConnections;
        }

        /// <summary>
        /// Stop accepting new connections.
        /// </summary>
        public void Stop()
        {
            if (!_HttpListener.IsListening)
            {
                throw new InvalidOperationException("EonaCat Webserver is already stopped.");
            }

            if (_HttpListener != null && _HttpListener.IsListening)
            {
                _HttpListener.Stop();
            }

            if (_TokenSource != null && !_TokenSource.IsCancellationRequested)
            {
                _TokenSource.Cancel();
            }
        }

        /// <summary>
        /// Tear down the server and dispose of background workers.
        /// Do not use this object after disposal.
        /// </summary>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (_HttpListener != null && _HttpListener.IsListening)
                {
                    Stop();

                    _HttpListener.Close();
                }

                Events.HandleServerDisposing(this, EventArgs.Empty);

                _HttpListener = null;
                _Settings = null;
                _Routes = null;
                _TokenSource = null;
                _AcceptConnections = null;

                Events = null;
                Statistics = null;
            }
        }

        private void LoadRoutes()
        {
            var staticRoutes = _Assembly
                .GetTypes() // Get all classes from assembly
                .SelectMany(x => x.GetMethods()) // Get all methods from assembly
                .Where(IsStaticRoute); // Only select methods that are valid routes

            var parameterRoutes = _Assembly
                .GetTypes() // Get all classes from assembly
                .SelectMany(x => x.GetMethods()) // Get all methods from assembly
                .Where(IsParameterRoute); // Only select methods that are valid routes

            var dynamicRoutes = _Assembly
                .GetTypes() // Get all classes from assembly
                .SelectMany(x => x.GetMethods()) // Get all methods from assembly
                .Where(IsDynamicRoute); // Only select methods that are valid routes

            foreach (var staticRoute in staticRoutes)
            {
                var attribute = staticRoute.GetCustomAttributes().OfType<StaticRouteAttribute>().First();
                if (!_Routes.Static.Exists(attribute.Method, attribute.Path))
                {
                    Events.Logger?.Invoke(_Header + "adding static route " + attribute.Method.ToString() + " " + attribute.Path);
                    _Routes.Static.Add(attribute.Method, attribute.Path, ToRouteMethod(staticRoute), attribute.GUID, attribute.Metadata);
                }
            }

            foreach (var parameterRoute in parameterRoutes)
            {
                var attribute = parameterRoute.GetCustomAttributes().OfType<ParameterRouteAttribute>().First();
                if (!_Routes.Parameter.Exists(attribute.Method, attribute.Path))
                {
                    Events.Logger?.Invoke(_Header + "adding parameter route " + attribute.Method.ToString() + " " + attribute.Path);
                    _Routes.Parameter.Add(attribute.Method, attribute.Path, ToRouteMethod(parameterRoute), attribute.GUID, attribute.Metadata);
                }
            }

            foreach (var dynamicRoute in dynamicRoutes)
            {
                var attribute = dynamicRoute.GetCustomAttributes().OfType<DynamicRouteAttribute>().First();
                if (!_Routes.Dynamic.Exists(attribute.Method, attribute.Path))
                {
                    Events.Logger?.Invoke(_Header + "adding dynamic route " + attribute.Method.ToString() + " " + attribute.Path);
                    _Routes.Dynamic.Add(attribute.Method, attribute.Path, ToRouteMethod(dynamicRoute), attribute.GUID, attribute.Metadata);
                }
            }
        }

        private bool IsStaticRoute(MethodInfo method)
        {
            return method.GetCustomAttributes().OfType<StaticRouteAttribute>().Any()
               && method.ReturnType == typeof(Task)
               && method.GetParameters().Length == 1
               && method.GetParameters().First().ParameterType == typeof(HttpContext);
        }

        private bool IsParameterRoute(MethodInfo method)
        {
            return method.GetCustomAttributes().OfType<ParameterRouteAttribute>().Any()
               && method.ReturnType == typeof(Task)
               && method.GetParameters().Length == 1
               && method.GetParameters().First().ParameterType == typeof(HttpContext);
        }

        private bool IsDynamicRoute(MethodInfo method)
        {
            return method.GetCustomAttributes().OfType<DynamicRouteAttribute>().Any()
               && method.ReturnType == typeof(Task)
               && method.GetParameters().Length == 1
               && method.GetParameters().First().ParameterType == typeof(HttpContext);
        }

        private Func<HttpContext, Task> ToRouteMethod(MethodInfo method)
        {
            if (method.IsStatic)
            {
                return (Func<HttpContext, Task>)Delegate.CreateDelegate(typeof(Func<HttpContext, Task>), method);
            }
            else
            {
                object instance = Activator.CreateInstance(method.DeclaringType ?? throw new Exception("EonaCat Network: Declaring class is null"));
                return (Func<HttpContext, Task>)Delegate.CreateDelegate(typeof(Func<HttpContext, Task>), instance, method);
            }
        }

        private async Task AcceptConnections(CancellationToken token)
        {
            try
            {
                while (_HttpListener.IsListening)
                {
                    if (_RequestCount >= _Settings.IO.MaxRequests)
                    {
                        await Task.Delay(100, token).ConfigureAwait(false);
                        continue;
                    }

                    System.Net.HttpListenerContext listenercontext = await _HttpListener.GetContextAsync().ConfigureAwait(false);
                    Interlocked.Increment(ref _RequestCount);
                    HttpContext context = null;

                    Task unawaited = Task.Run(async () =>
                    {
                        DateTime startTime = DateTime.Now;

                        try
                        {
                            Events.HandleConnectionReceived(this, new ConnectionEventArgs(
                                listenercontext.Request.RemoteEndPoint.Address.ToString(),
                                listenercontext.Request.RemoteEndPoint.Port));

                            context = new HttpContext(listenercontext, _Settings, Events);

                            Events.HandleRequestReceived(this, new RequestEventArgs(context));

                            if (_Settings.Debug.Requests)
                            {
                                Events.Logger?.Invoke(
                                    _Header + context.Request.Source.IpAddress + ":" + context.Request.Source.Port + " " +
                                    context.Request.Method.ToString() + " " + context.Request.Url.RawWithoutQuery);
                            }

                            Statistics.IncrementRequestCounter(context.Request.Method);
                            Statistics.IncrementReceivedPayloadBytes(context.Request.ContentLength);

                            if (!_Settings.AccessControl.Permit(context.Request.Source.IpAddress))
                            {
                                Events.HandleRequestDenied(this, new RequestEventArgs(context));

                                if (_Settings.Debug.AccessControl)
                                {
                                    Events.Logger?.Invoke(_Header + context.Request.Source.IpAddress + ":" + context.Request.Source.Port + " denied due to access control");
                                }

                                listenercontext.Response.StatusCode = 403;
                                listenercontext.Response.Close();
                                return;
                            }

                            if (context.Request.Method == HttpMethod.OPTIONS)
                            {
                                if (_Routes.Preflight != null)
                                {
                                    if (_Settings.Debug.Routing)
                                    {
                                        Events.Logger?.Invoke(
                                            _Header + "preflight route for " + context.Request.Source.IpAddress + ":" + context.Request.Source.Port + " " +
                                            context.Request.Method.ToString() + " " + context.Request.Url.RawWithoutQuery);
                                    }

                                    await _Routes.Preflight(context).ConfigureAwait(false);
                                    return;
                                }
                            }

                            bool terminate = false;
                            if (_Routes.PreRouting != null)
                            {
                                terminate = await _Routes.PreRouting(context).ConfigureAwait(false);
                                if (terminate)
                                {
                                    if (_Settings.Debug.Routing)
                                    {
                                        Events.Logger?.Invoke(
                                            _Header + "prerouting terminated connection for " + context.Request.Source.IpAddress + ":" + context.Request.Source.Port + " " +
                                            context.Request.Method.ToString() + " " + context.Request.Url.RawWithoutQuery);
                                    }

                                    return;
                                }
                            }

                            if (context.Request.Method == HttpMethod.GET || context.Request.Method == HttpMethod.HEAD)
                            {
                                if (_Routes.Content.Match(context.Request.Url.RawWithoutQuery, out ContentRoute cr))
                                {
                                    if (_Settings.Debug.Routing)
                                    {
                                        Events.Logger?.Invoke(
                                            _Header + "content route for " + context.Request.Source.IpAddress + ":" + context.Request.Source.Port + " " +
                                            context.Request.Method.ToString() + " " + context.Request.Url.RawWithoutQuery);
                                    }

                                    context.RouteType = RouteTypeEnum.Content;
                                    context.Route = cr;
                                    await _Routes.ContentHandler.Process(context, token).ConfigureAwait(false);
                                    return;
                                }
                            }

                            Func<HttpContext, Task> handler = _Routes.Static.Match(context.Request.Method, context.Request.Url.RawWithoutQuery, out StaticRoute sr);
                            if (handler != null)
                            {
                                if (_Settings.Debug.Routing)
                                {
                                    Events.Logger?.Invoke(
                                        _Header + "static route for " + context.Request.Source.IpAddress + ":" + context.Request.Source.Port + " " +
                                        context.Request.Method.ToString() + " " + context.Request.Url.RawWithoutQuery);
                                }

                                context.RouteType = RouteTypeEnum.Static;
                                context.Route = sr;
                                await handler(context).ConfigureAwait(false);
                                return;
                            }

                            handler = _Routes.Parameter.Match(context.Request.Method, context.Request.Url.RawWithoutQuery, out Dictionary<string, string> parameters, out ParameterRoute pr);
                            if (handler != null)
                            {
                                context.Request.Url.Parameters = new Dictionary<string, string>(parameters);

                                if (_Settings.Debug.Routing)
                                {
                                    Events.Logger?.Invoke(
                                        _Header + "parameter route for " + context.Request.Source.IpAddress + ":" + context.Request.Source.Port + " " +
                                        context.Request.Method.ToString() + " " + context.Request.Url.RawWithoutQuery);
                                }

                                context.RouteType = RouteTypeEnum.Parameter;
                                context.Route = pr;
                                await handler(context).ConfigureAwait(false);
                                return;
                            }

                            handler = _Routes.Dynamic.Match(context.Request.Method, context.Request.Url.RawWithoutQuery, out DynamicRoute dr);
                            if (handler != null)
                            {
                                if (_Settings.Debug.Routing)
                                {
                                    Events.Logger?.Invoke(
                                        _Header + "dynamic route for " + context.Request.Source.IpAddress + ":" + context.Request.Source.Port + " " +
                                        context.Request.Method.ToString() + " " + context.Request.Url.RawWithoutQuery);
                                }

                                context.RouteType = RouteTypeEnum.Dynamic;
                                context.Route = dr;
                                await handler(context).ConfigureAwait(false);
                                return;
                            }

                            if (_Routes.Default != null)
                            {
                                if (_Settings.Debug.Routing)
                                {
                                    Events.Logger?.Invoke(
                                        _Header + "default route for " + context.Request.Source.IpAddress + ":" + context.Request.Source.Port + " " +
                                        context.Request.Method.ToString() + " " + context.Request.Url.RawWithoutQuery);
                                }

                                context.RouteType = RouteTypeEnum.Default;
                                await _Routes.Default(context).ConfigureAwait(false);
                                return;
                            }
                            else
                            {
                                if (_Settings.Debug.Routing)
                                {
                                    Events.Logger?.Invoke(
                                        _Header + "default route not found for " + context.Request.Source.IpAddress + ":" + context.Request.Source.Port + " " +
                                        context.Request.Method.ToString() + " " + context.Request.Url.RawWithoutQuery);
                                }

                                context.Response.StatusCode = 404;
                                context.Response.ContentType = Pages.Default404Page.ContentType;
                                await context.Response.Send(Pages.Default404Page.Content).ConfigureAwait(false);
                                return;
                            }
                        }
                        catch (Exception eInner)
                        {
                            context.Response.StatusCode = 500;
                            context.Response.ContentType = Pages.Default500Page.ContentType;
                            await context.Response.Send(Pages.Default500Page.Content).ConfigureAwait(false);
                            Events.HandleExceptionEncountered(this, new ExceptionEventArgs(context, eInner));
                        }
                        finally
                        {
                            Interlocked.Decrement(ref _RequestCount);

                            if (context != null && context.Response != null && context.Response.ResponseSent)
                            {
                                Events.HandleResponseSent(this, new ResponseEventArgs(context, TotalMsFrom(startTime)));
                                Statistics.IncrementSentPayloadBytes(context.Response.ContentLength);
                            }
                        }
                    }, token);
                }
            }
            catch (TaskCanceledException)
            {
            }
            catch (OperationCanceledException)
            {
            }
            catch (HttpListenerException)
            {
            }
            catch (Exception e)
            {
                Events.HandleExceptionEncountered(this, new ExceptionEventArgs(null, e));
            }
            finally
            {
                Events.HandleServerStopped(this, EventArgs.Empty);
            }
        }

        private double TotalMsFrom(DateTime startTime)
        {
            try
            {
                DateTime endTime = DateTime.Now;
                TimeSpan totalTime = (endTime - startTime);
                return totalTime.TotalMilliseconds;
            }
            catch (Exception)
            {
                return -1;
            }
        }
    }
}