// This file is part of the EonaCat project(s) which is released under the Apache License.
// See the LICENSE file or go to https://EonaCat.com/License for full license details.

using System;
using System.Collections;
using System.Collections.Generic;

namespace EonaCat.Network
{
    internal class Payload : IEnumerable<byte>
    {
        private ushort _code;
        private bool _codeSet;
        private readonly byte[] _data;
        private readonly long _length;
        private string _reason;
        private bool _reasonSet;

        public static readonly Payload Empty;

        public static readonly ulong MaxLength;

        static Payload()
        {
            Empty = new Payload();
            MaxLength = long.MaxValue;
        }

        internal Payload()
        {
            _code = (ushort)CloseStatusCode.NoStatus;
            _reason = string.Empty;

            _data = WSClient.EmptyBytes;

            _codeSet = true;
            _reasonSet = true;
        }

        internal Payload(Payload original)
        {
            _code = original._code;
            _codeSet = original._codeSet;
            ExtensionDataLength = original.ExtensionDataLength;
            _length = original._length;
            _reason = original._reason;
            _reasonSet = original._reasonSet;

            _data = new byte[_length];
            original._data.CopyTo(_data, 0);
        }

        internal Payload(byte[] data)
          : this(data, data.LongLength)
        {
        }

        internal Payload(byte[] data, long length)
        {
            _data = data;
            _length = length;
        }

        internal Payload(ushort code, string reason)
        {
            _code = code;
            _reason = reason ?? string.Empty;

            _data = code.Append(reason);
            _length = _data.LongLength;

            _codeSet = true;
            _reasonSet = true;
        }

        internal ushort Code
        {
            get
            {
                if (!_codeSet)
                {
                    _code = _length > 1
                            ? _data.SubArray(0, 2).ToUInt16(ByteOrder.Big)
                            : (ushort)CloseStatusCode.NoStatus;

                    _codeSet = true;
                }

                return _code;
            }
        }

        internal long ExtensionDataLength { get; set; }

        internal bool HasReservedCode => _length > 1 && Code.IsReserved();

        internal string Reason
        {
            get
            {
                if (!_reasonSet)
                {
                    _reason = _length > 2
                              ? _data.SubArray(2, _length - 2).UTF8Decode()
                              : string.Empty;

                    _reasonSet = true;
                }

                return _reason;
            }
        }

        public byte[] ApplicationData => ExtensionDataLength > 0
                       ? _data.SubArray(ExtensionDataLength, _length - ExtensionDataLength)
                       : _data;

        public byte[] ExtensionData => ExtensionDataLength > 0
                       ? _data.SubArray(0, ExtensionDataLength)
                       : WSClient.EmptyBytes;

        public ulong Length => (ulong)_length;

        internal void Mask(byte[] key)
        {
            for (long i = 0; i < _length; i++)
            {
                _data[i] = (byte)(_data[i] ^ key[i % 4]);
            }
        }

        public IEnumerator<byte> GetEnumerator()
        {
            foreach (var b in _data)
            {
                yield return b;
            }
        }

        public byte[] ToArray()
        {
            return _data;
        }

        public override string ToString()
        {
            return BitConverter.ToString(_data);
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }
    }
}